#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Recording utilities.
"""

__all__ = ["record_quantization_variables"]

import numpy as np
import tensorflow as tf

from .transforms.transforms_utils import get_layers_by_type
from ..layers import recording, StatefulRecurrent, reset_states


def record_quantization_variables(model):
    """Helper method to record quantization objects in the graph.

    Passing a dummy sample through the model in recording mode, this triggers the
    recording of all dynamic quantization objects.

    Args:
        model (keras.Model): model for which objects need to be recorded.
    """
    def _gen_dummy_sample(shape, type=np.float32):
        if issubclass(type, np.floating):
            sample = np.random.randint(0, 255, size=(1, *shape))
            return sample.astype(type)
        return np.random.randint(np.iinfo(type).min, np.iinfo(type).max,
                                 size=(1, *shape), dtype=type)

    recurrent_layers = get_layers_by_type(model, StatefulRecurrent)

    with recording(True):
        if len(recurrent_layers):
            # Reset model (states and counter) to ensure a proprer recording
            reset_states(model)
            # Build a tf.function to run in graph mode
            model_func = tf.function(model)
            # The number of samples that will allow proper recording is equal to the lowest
            # common multiple among the subsampling ratios found in the model.
            sub_sample_ratios = []
            for rec in recurrent_layers:
                sub_sample_ratios.append(rec.subsample_ratio)
            num_samples = 2 * np.lcm.reduce(sub_sample_ratios)
            # For recurrent models, inputs are expected to be int16 so random samples are generated
            # accordingly
            sample = _gen_dummy_sample((num_samples, ) + model.input.shape[1:], np.int16)
            # Custom loop to ensure counter and subsampling ratio condition are met
            for i in range(num_samples):
                model_func(sample[:, i, :])
            # Reset model again
            reset_states(model)
        else:
            # Create sample and pass it through the model to calibrate variables
            sample = _gen_dummy_sample(model.input.shape[1:])
            model(sample)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["QuantizedConv2D", "get_qconv"]

import numpy as np

from onnx import AttributeProto as AP, TensorProto as TP
from onnx.helper import make_node

from .base_layer import OnnxLayer, register_node_format
from .subgraph_ops import cast_tensors_to, get_pool_ops, get_scale_out_ops
from .subgraph_ops.activation import get_activation_ops
from .subgraph_ops.padding import get_padding_ops, transform_pads_into_array
from .compute_shapes import compute_onnx_conv_output
from .layer_compatibility import check_clip_relu_compatibility
from ..graph_tools import (TENSOR_SHAPE, get_field, get_node, get_variable, get_tensor_shape,
                           check_node_attributes)
from ..quantization.weights import quantize_weights, quantize_vector, fold_zero_point, align_to
from ..quantization.outputs import downscale


def get_qconv(nodes, graph):
    conv_node = nodes[0]
    assert conv_node.op_type == 'Conv'

    # Check supported attributes
    weights = get_variable(conv_node.input[1], graph)
    valid_attr = {'auto_pad': ['NOTSET'], 'dilations': [[1] * (weights.ndim - 2)], 'group': [1]}
    check_node_attributes(conv_node, valid_attr)

    # Retrieve attributes
    strides = get_field(conv_node, 'strides', (1, 1))
    pool_type = "none"
    pool_size = (2, 2)
    pool_strides = (1, 1)
    pool_node = get_node(nodes, 'MaxPool')
    pool_pads = [0, 0, 0, 0]
    if pool_node:
        pool_type = "max"
        # kernel_shape attribute is mandatory for MaxPool
        pool_size = get_field(pool_node, 'kernel_shape')
        pool_strides = get_field(pool_node, 'strides', pool_strides)
        pool_pads = get_field(pool_node, "pads", pool_pads)
    pool_node = get_node(nodes, 'GlobalAveragePool')
    if pool_node:
        pool_type = "gap"

    act_node = get_node(nodes, 'Relu')
    clip_node = get_node(nodes, 'Clip')

    qconv = QuantizedConv2D(strides=strides,
                            pool_type=pool_type,
                            pool_size=pool_size,
                            pool_strides=pool_strides,
                            pool_pads=pool_pads,
                            activation=bool(act_node) or bool(clip_node),
                            name=conv_node.name)

    # Sets the weights to configure the operation chain
    qconv.set_weight("kernel", weights)
    # If third attribute is there and it is not empty, then there is a bias
    if len(conv_node.input) == 3 and conv_node.input[2]:
        qconv.set_weight("bias", get_variable(conv_node.input[2], graph))
    pads = get_field(conv_node, 'pads', False)
    if pads:
        qconv.set_weight("pads", transform_pads_into_array(pads))

    if clip_node:
        check_clip_relu_compatibility(clip_node, graph)
        qconv.set_weight("max_value", get_variable(clip_node.input[2], graph))

    return qconv


@register_node_format(requires_downscale=True)
class QuantizedConv2D(OnnxLayer):
    """Intermediate representation of QLinearConv() + MaxPool() + ReLU() as an exportable node.

    Args:
        strides (list of int, optional): the convolutional strides. Defaults to [1, 1].
        pool_type (str, optional): the pool type, one of {"none", "max", "gap"}. Defaults to "none".
        pool_size (list of int, optional): the kernel pool shape.
            Ignore it when pool_type != "max". Defaults to (2, 2).
        pool_stride (list of int, optional): the kernel strides.
            Ignore it when pool_type != "max". Defaults to (2, 2).
        pool_pads (list of int, optional): the size of each padding dimension.
            Ignore it when pool_type != "max". Defaults to [0, 0, 0, 0].
        input_conv (bool, optional): whether it is extended the set of operations of
            the basic QuantizedConv2D, allowing to modify the padding value per input channel.
            Defaults to False.
        activation (bool, optional): whether to apply relu operation. Defaults to False.
        name (str, optional): the node name. Defaults to ''.
    """

    def __init__(self,
                 strides=[1, 1],
                 pool_type="none",
                 pool_size=(2, 2),
                 pool_strides=(2, 2),
                 pool_pads=[0, 0, 0, 0],
                 activation=False,
                 name=''):
        assert pool_type in ["none", "max", "gap"]
        super().__init__("QuantizedConv2D",
                         strides=strides,
                         pool_size=pool_size,
                         pool_strides=pool_strides,
                         pool_pads=pool_pads,
                         name=name)

        # Save properties need to serialize operation name
        self.serialize_attr["pool_type"] = pool_type
        self.serialize_attr["activation"] = activation
        self.serialize_attr["scale"] = True

        # Declare weights
        self._add_weight("kernel")
        self._add_weight("bias")
        self._add_weight("max_value")
        self._add_weight("pads", dtype="int64")

    def __build__(self, input_ts, downscale=True):
        assert input_ts.dtype in (np.uint8, np.int8)
        assert downscale, f"{self.name} ({self.base_name}) does not support 32bit output"
        assert self.weights["kernel"].ndim == 4

        # The chain of operations is modified by the type of input:
        kernel_shape = self.weights["kernel"].shape
        if input_ts.dtype == np.uint8:
            self.base_name = "QuantizedInputConv2D"
            if self.weights["bias"].size == 0:
                # Bias is mandatory on this configuration
                filters = kernel_shape[0]
                self.set_weight("bias", np.zeros(filters, dtype="float32"))

        # Initialize weights
        if self.weights["pads"].size == 0:
            self.set_weight("pads", np.zeros(len(kernel_shape) * 2, dtype="int64"))

        # Compute output shape
        conv_output_shape = compute_onnx_conv_output(self, input_ts.shape)
        output_ts = TENSOR_SHAPE(conv_output_shape, np.dtype("int8"))
        return output_ts

    def __quantize__(self, qinput, out_tensor_range, force_fp=False):
        i_scale = qinput.weights["scale"]

        # Perform cross-layer equalization, i.e.: rescale weights with input scale.
        # To do that first reshape i_scale to put it into axis = 1 and be capable of broadcasting.
        assert i_scale.ndim <= 1
        kernel = self.weights["kernel"]
        kernel = kernel / align_to(i_scale, kernel.ndim, axis=1)
        # Quantize and set weights
        qweights, i_scale = quantize_weights(kernel)

        # Prepare tensors list with unique names
        conv_name = self.name
        prefix = conv_name + "_"
        weights_dict = {}
        bias = self.weights["bias"]
        if "InputConv" in self.op_type:
            # If calibration was done per tensor, repeat zero point over each channel
            zero_point = qinput.weights["zero_point"]
            if zero_point.size == 1:
                zero_point = np.repeat(zero_point, kernel.shape[1])
            weights_dict[prefix + "Xpad"] = zero_point
            # Fold zero point in bias
            # Note: Dequantize kernel instead to use the float to reduce quantization error
            kernel = qweights / align_to(i_scale, qweights.ndim, axis=0)
            bias = fold_zero_point(bias, kernel, zero_point)
        weights_dict[prefix + "Wi"] = qweights
        if "Biased" in self.op_type:
            qbias = quantize_vector(bias, i_scale)
            weights_dict[prefix + "B"] = qbias
        weights_dict[prefix + "pads"] = self.weights["pads"]

        # Reshape i_scale to match with channel axis
        i_scale = align_to(i_scale, qweights.ndim)

        # Quantize max value when there is an activation
        if "Clipped" in self.op_type:
            qmax_value = quantize_vector(self.weights["max_value"], i_scale, signed=False)
            weights_dict[prefix + "max_value"] = qmax_value

        # Fold spatial dimension when GAP
        if "GlobalAvgPool" in self.op_type:
            input_shape = get_tensor_shape(self.input)
            input_shape = compute_onnx_conv_output(self, input_shape, apply_pool=False)
            i_scale *= input_shape[-2] * input_shape[-1]

        # Now consider calibrated output range
        scale, s_out, ocalib_scale = downscale(out_tensor_range, i_scale, force_fp=force_fp)
        weights_dict.update({prefix + "M": scale.astype("uint8"), prefix + "S_out": s_out})

        # Return quantized weights and ouput scale
        return weights_dict, ocalib_scale

    @staticmethod
    def build_subgraph(op_type):
        # Cast input, weights (and bias) into float.
        t_names = ["X", "", "W", ""]
        if "InputConv" in op_type:
            t_names[1] = "x_pad_value"
        if "Biased" in op_type:
            t_names[-1] = "bias"
        nodes, t_names = cast_tensors_to(t_names)

        # Pad + convolution
        nodes += get_padding_ops(t_names[0], "Xi", t_names[1])
        conv_tensor_names = nodes[-1].output[:1] + t_names[2:]
        nodes.append(make_node("Conv", inputs=conv_tensor_names, outputs=["Yi"]))
        nodes[-1].attribute.append(AP(name="strides", ref_attr_name="strides", type=AP.INTS))

        # Maxpool (optional)
        if "MaxPool" in op_type:
            # Replace previous output as maxpool input
            nodes[-1].output.__setitem__(0, nodes[-1].op_type)
            nodes += get_pool_ops(nodes[-1].output[0], "Yi", pool_op_type="MaxPool")

        # Activation (optional)
        if "ReLU" in op_type:
            # Replace previous output as relu input
            nodes[-1].output.__setitem__(0, nodes[-1].op_type)
            nodes += get_activation_ops(nodes[-1].output[0], "Yi", "ReLUClipped" in op_type)

        # AvgPool (optional)
        if "GlobalAvgPool" in op_type:
            # Replace previous output as maxpool input
            nodes[-1].output.__setitem__(0, nodes[-1].op_type)
            nodes += get_pool_ops(nodes[-1].output[0], "Yi", pool_op_type="GlobalAvgPool")

        # Scale out (with saturation) in float domain
        shift_nodes, shift_t_names = cast_tensors_to(["Scale", "Shift"])
        nodes += shift_nodes
        nodes += get_scale_out_ops("Yi", "Yscaled", *shift_t_names)
        # Cast output to expect type
        nodes.append(make_node("Cast", ["Yscaled"], ["Y"], to=TP.INT8))
        return nodes



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["recording", "TensorRecorder", "FixedPointRecorder", "QFloatRecorder", "Recorder",
           "NonTrackVariable", "NonTrackFixedPointVariable"]

import os
import tensorflow as tf
from contextlib import contextmanager

from ..tensors import FixedPoint, QFloat

RECORDING_ENV = "RECORDING_ENABLED"


@contextmanager
def recording(enable):
    """Enable or disable recording.

    Args:
        enable (bool): True to enable recording, False to disable it
    """
    value = "1" if enable else "0"
    _prev_state = os.environ.get(RECORDING_ENV, None)
    try:
        os.environ[RECORDING_ENV] = value
        yield
    finally:
        # Recover default value
        if _prev_state is not None:
            os.environ[RECORDING_ENV] = _prev_state
        else:
            os.environ.pop(RECORDING_ENV)


class BaseNonTrackVariable():
    """Base interface class for temporary variables that should be tracked only during the call and
        which does not require to be serialized within the layer.
    """

    def __init__(self, name=""):
        self._name = name
        self._var = None

    @property
    def var(self):
        raise NotImplementedError

    @tf.function
    def set_var(self, new_value):
        raise NotImplementedError

    @tf.function
    def init_var(self):
        raise NotImplementedError

    def reset_var(self):
        raise NotImplementedError


class NonTrackVariable(BaseNonTrackVariable):
    """A wrapper class for the temporary Tensor variables that should be tracked only during the
        call and which does not require to be serialized within the layer.
    """

    @property
    def var(self):
        return self._var

    @tf.function
    def set_var(self, new_value):
        self._var.assign(new_value)

    @tf.function
    def init_var(self, init_value, validate_shape=False):
        """Function that creates and initializes a variable, if it doesn't exist. This variable will
            be integrated in the layer graph and tracked (but not within the main layer variables).
            See pattern defined here: https://www.tensorflow.org/guide/function#creating_tfvariables

        Args:
            init_value (tf.Tensor): Tensor, or Python object convertible to a Tensor which is the
                initial value for the Variable. The initial value must have a shape specified
                unless validate_shape is set to False.
            validate_shape (bool, optional): If False, allows the variable to be initialized with a
                value of unknown shape. If True the shape of initial_value must be known.
                Defaults to False.
        """

        if self._var is None:
            self._var = tf.Variable(init_value, trainable=False,
                                    validate_shape=validate_shape,
                                    name=self._name,
                                    synchronization=tf.VariableSynchronization.ON_READ)

    def reset_var(self):
        """ Reset internal var."""
        if self._var is not None:
            self._var.assign_add(-self._var)


class NonTrackFixedPointVariable(BaseNonTrackVariable):
    """A wrapper class for the temporary FixedPoint variables that should be tracked only during
        the call and which does not require to be serialized within the layer.
    """

    def __init__(self, name=""):
        super().__init__(name)
        self._frac_bits = None
        self._value_bits = None

    @property
    def var(self):
        return None if self._var is None else FixedPoint(self._var, self._value_bits,
                                                         self._frac_bits)

    def promote(self, target_value_bits):
        return self.var.promote(target_value_bits)

    def expand(self, target_value_bits):
        return self.var.expand(target_value_bits)

    @tf.function
    def set_var(self, new_value):
        assert (isinstance(new_value, FixedPoint))
        self._var.assign(new_value.values)
        self._frac_bits.assign(new_value.frac_bits)
        self._value_bits = new_value.value_bits

    @tf.function
    def init_var(self, init_value, validate_shape=False):
        """Function that creates and initializes a variable, if it doesn't exist. This variable will
            be integrated in the layer graph and tracked (but not within the main layer variables).
            See pattern defined here: https://www.tensorflow.org/guide/function#creating_tfvariables

        Args:
            init_value (FixedPoint): initial FixedPoint variable.
            validate_shape (bool, optional): If False, allows the variable to be initialized with a
                value of unknown shape. If True the shape of initial_value must be known.
                Defaults to False.
        """
        assert (isinstance(init_value, FixedPoint))
        if self._var is None:
            self._var = tf.Variable(init_value.values, trainable=False,
                                    validate_shape=validate_shape,
                                    name=self._name,
                                    synchronization=tf.VariableSynchronization.ON_READ)

            self._frac_bits = tf.Variable(init_value.frac_bits, trainable=False,
                                          validate_shape=validate_shape,
                                          name=self._name + "/frac_bits",
                                          synchronization=tf.VariableSynchronization.ON_READ)

            self._value_bits = init_value.value_bits

    def reset_var(self):
        """ Reset internal var."""
        if self._var is not None:
            self._var.assign_add(-self._var)
            self._frac_bits.assign_add(-self._frac_bits)


class NonTrackQFloatVariable(BaseNonTrackVariable):
    """A wrapper class for the temporary QFloat variables that should be tracked only during
        the call and which does not require to be serialized within the layer.
    """

    def __init__(self, name=""):
        super().__init__(name)
        self._fp = NonTrackFixedPointVariable(name=name)
        self._scales = None

    @property
    def var(self):
        return None if self._scales is None else QFloat(self._fp.var, self._scales)

    def promote(self, target_value_bits):
        return self.var.promote(target_value_bits)

    def expand(self, target_value_bits):
        return self.var.expand(target_value_bits)

    @tf.function
    def set_var(self, new_value):
        assert (isinstance(new_value, QFloat))
        self._fp.set_var(new_value.fp)
        self._scales.assign(new_value.scales)

    @tf.function
    def init_var(self, init_value, validate_shape=False):
        """Function that creates and initializes a variable, if it doesn't exist. This variable will
            be integrated in the layer graph and tracked (but not within the main layer variables).
            See pattern defined here: https://www.tensorflow.org/guide/function#creating_tfvariables

        Args:
            init_value (QFloat): initial QFloat variable.
            validate_shape (bool, optional): If False, allows the variable to be initialized with a
                value of unknown shape. If True the shape of initial_value must be known.
                Defaults to False.
        """
        assert (isinstance(init_value, QFloat))
        if self._scales is None:
            self._fp.init_var(init_value.fp, validate_shape)

            self._scales = tf.Variable(init_value.scales, trainable=False,
                                       validate_shape=validate_shape,
                                       name=self._name + "/scales",
                                       synchronization=tf.VariableSynchronization.ON_READ)

    def reset_var(self):
        """ Reset internal var."""
        if self._scales is not None:
            self._fp.reset_var()
            self._scales.assign_add(-self._scales)


class Recorder():
    """A wrapper class with useful properties/methods for a recorder
    """

    def __init__(self, *args, name="", **kwargs):
        super().__init__(*args, **kwargs)
        self._name = name

    @property
    def value(self):
        """Get the recorded value.

        Returns:
            Any: value of the stored record or None.
        """
        raise NotImplementedError("Child must implement this property")

    @property
    def name(self):
        return self._name if self._name is not None else "record"

    @property
    def recording(self):
        """Flag to specify if the object is in recording mode or not.

        Returns:
            bool: True if recording mode is enabled, False otherwise.
        """
        value = os.environ.get(RECORDING_ENV, "0")
        return (value == "1")

    def __call__(self, inputs):
        raise NotImplementedError("Child must implement this function")


class TensorRecorder(Recorder, NonTrackVariable):
    """Wrapper class to store and retrieve a tf.Tensor extracted from a graph.

    This is mainly used to recover FixedPoint alignment shift information.
    """

    @property
    def value(self):
        """Get the recorded value.

        Returns:
            tf.Tensor: value of the stored record or None.
        """
        return None if self._var is None else self._var.value()

    def __call__(self, inputs):
        """Record the values of the inputs if recording is True.

        Args:
            inputs (tf.Tensor): new values.

        Returns:
            tf.Tensor: the inputs.
        """
        self.init_var(tf.zeros_like(inputs), True)
        if self.recording:
            # Store the new values
            self.set_var(inputs)
        return inputs


class FixedPointRecorder(Recorder):
    """Wrapper class to store and retrieve a FixedPoint extracted from a graph.

    This is mainly used to recover FixedPoint quantized weights.
    """

    def __init__(self, name=""):
        super().__init__(name=name)
        if name != "":
            name += "/"
        self._values = TensorRecorder(name + "values/record")
        self._frac_bits = TensorRecorder(name + "frac_bits/record")
        self._value_bits = None

    @property
    def value(self):
        """Get the recorded value.

        Returns:
            :obj:`FixedPoint`: value of the stored record or None.
        """
        return None if self._value_bits is None else FixedPoint(self._values.value,
                                                                self._value_bits,
                                                                self._frac_bits.value)

    def __call__(self, inputs):
        """Record the values of the inputs if recording is True.

        Args:
            inputs (:obj:`FixedPoint`): new values.

        Returns:
            :obj:`FixedPoint`: the inputs.
        """
        if self.recording:
            self._value_bits = inputs.value_bits
        self._values(inputs.values)
        self._frac_bits(inputs.frac_bits)
        return inputs


class QFloatRecorder(Recorder):
    """Wrapper class to store and retrieve a QFloat extracted from a graph.

    This is mainly used to recover QFloat quantized weights.
    """

    def __init__(self, name=""):
        super().__init__(name=name)
        self._fp = FixedPointRecorder(name)
        scales_name = "scales/record"
        if name != "":
            scales_name = name + "/" + scales_name
        self._scales = TensorRecorder(scales_name)

    @property
    def value(self):
        """Get the recorded value.

        Returns:
            :obj:`QFloat`: value of the stored record or None.
        """
        return None if self._fp.value is None else QFloat(self._fp.value, self._scales.value)

    def __call__(self, inputs):
        """Record the values of the inputs if recording is True.

        Args:
            inputs (:obj:`QFloat`): new values.

        Returns:
            :obj:`QFloat`: the inputs.
        """
        self._fp(inputs.fp)
        self._scales(inputs.scales)
        return inputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["QuantizedDense"]

import tensorflow as tf
import keras

from .layers_base import (register_quantize_target, rescale_outputs, tensor_inputs,
                          neural_layer_init, register_aligned_inputs, QuantizedLayer)
from ..tensors import QTensor


@register_quantize_target(keras.layers.Dense)
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedDense(QuantizedLayer, keras.layers.Dense):
    """A Dense layer that operates on quantized inputs and weights

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """
    @neural_layer_init(False)
    def __init__(self, *args, quant_config=None, **kwargs):
        # Limit buffer bitwidth to 27 for HW constraint
        self.quant_config['buffer_bitwidth'] = min(28, self.quant_config['buffer_bitwidth'])
        self.buffer_bitwidth = self.quant_config['buffer_bitwidth'] - 1

    @tensor_inputs([QTensor, tf.Tensor])
    @rescale_outputs
    def call(self, inputs):
        # Quantize the weights
        kernel = self.weight_quantizer(self.kernel)

        outputs = tf.matmul(inputs, kernel)

        if self.use_bias:
            # Quantize and align biases
            bias = self.bias_quantizer(self.bias, outputs)
            outputs = tf.add(outputs, bias)

        return outputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["custom_pattern_scope"]

from collections import namedtuple
from inspect import signature
from contextlib import contextmanager
from .. import layers as onnx_qlayers

# Define named tuples for QuantizerPattern
QuantizePattern = namedtuple('QuantizerPattern', ['pattern', 'f'])

# List of supported patterns, together with matching function
CUSTOM_PATTERNS_MAP = []
PATTERNS_MAP = [
    QuantizePattern(("Conv", "Relu", "GlobalAveragePool"), [onnx_qlayers.get_qconv]),
    QuantizePattern(("Conv", "Relu", "MaxPool"), [onnx_qlayers.get_qconv]),
    QuantizePattern(("Conv", "GlobalAveragePool"), [onnx_qlayers.get_qconv]),
    QuantizePattern(("Conv", "Relu"), [onnx_qlayers.get_qdepthwise, onnx_qlayers.get_qconv]),
    QuantizePattern(("Conv", "Clip", "GlobalAveragePool"), [onnx_qlayers.get_qconv]),
    QuantizePattern(("Conv", "Clip", "MaxPool"), [onnx_qlayers.get_qconv]),
    QuantizePattern(("Conv", "Clip"), [onnx_qlayers.get_qdepthwise, onnx_qlayers.get_qconv]),
    QuantizePattern(("Conv",), [onnx_qlayers.get_qdepthwise, onnx_qlayers.get_qconv]),
    QuantizePattern(("Flatten", "Gemm", "Relu"), [onnx_qlayers.get_qgemm]),
    QuantizePattern(("Flatten", "Gemm", "Clip"), [onnx_qlayers.get_qgemm]),
    QuantizePattern(("Flatten", "Gemm"), [onnx_qlayers.get_qgemm]),
    QuantizePattern(("Gemm", "Relu"), [onnx_qlayers.get_qgemm]),
    QuantizePattern(("Gemm", "Clip"), [onnx_qlayers.get_qgemm]),
    QuantizePattern(("Gemm",), [onnx_qlayers.get_qgemm]),
    QuantizePattern(("Add", "Relu"), [onnx_qlayers.get_qadd]),
    QuantizePattern(("Add",), [onnx_qlayers.get_qadd]),
    QuantizePattern(("Concat",), [onnx_qlayers.get_qconcat]),
    QuantizePattern(("ConvTranspose", "Clip"), [onnx_qlayers.get_qconv_transpose]),
    QuantizePattern(("ConvTranspose", "Relu"), [onnx_qlayers.get_qconv_transpose]),
    QuantizePattern(("ConvTranspose",), [onnx_qlayers.get_qconv_transpose]),
    QuantizePattern(("Transpose", "Mul", "Add"), [onnx_qlayers.get_input_quantizer]),
    QuantizePattern(("Transpose", "Mul"), [onnx_qlayers.get_input_quantizer]),
    QuantizePattern(("Mul", "Add"), [onnx_qlayers.get_input_quantizer]),
    QuantizePattern(("Mul",), [onnx_qlayers.get_input_quantizer]),
]


@contextmanager
def custom_pattern_scope(new_patterns):
    """Register a custom pattern in the context to be used at quantization time.

    A pattern is understood as a sequence of continuous operations in the graph,
    whose representation can converge in an ``OnnxLayer``.

    Args:
        new_patterns (dict): a list of sequence of nodes (keys) and their mapper function (values).
    """
    # Use of global parameters
    global CUSTOM_PATTERNS_MAP
    # Transform input patterns in a valid format
    qpatterns = []
    for new_pattern, func in new_patterns.items():
        qpatterns.append(_custom_pattern_to_qpattern(new_pattern, func))
    try:
        # Extend CUSTOM_PATTERNS_MAP with new qpatterns
        CUSTOM_PATTERNS_MAP.extend(qpatterns)
        yield
    finally:
        # Restore to previous state
        CUSTOM_PATTERNS_MAP.clear()


def _custom_pattern_to_qpattern(pattern, func):
    assert callable(func), f"function has to be a callable. Receives: {func}"
    if len(signature(func).parameters) != 2:
        raise RuntimeError("function must have two inputs: sequence_nodes and graph")
    if isinstance(pattern, str):
        pattern = (pattern,)
    if not (isinstance(pattern, tuple) and all(isinstance(x, str) for x in pattern)):
        raise ValueError(f"Pattern must be a string-tuple. Receives: {pattern}")
    return QuantizePattern(pattern, [func])



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Transforms utility methods.
"""

from copy import deepcopy

from keras.models import Sequential
from keras.layers import Layer


def get_layers(config, layer_names):
    """Extracts layers from a model configuration.

    Args:
        config (dict): JSON formatted model configuration
        layer_names (list): list of layer names to extract

    Returns:
        list: layers configurations
    """
    return [layer for layer in config['layers'] if layer['config']['name'] in layer_names]


def get_layer_index(layers, layer_name):
    """Retrieves the layer index within the layer list.

    Args:
        layers (list): list of JSON formatted layers configurations
        layer_name (str): layer name to retrieve

    Returns:
        int: the layer index
    """
    for index, ly in enumerate(layers):
        if ly['config']['name'] == layer_name:
            return index
    return -1


def inbound_node_generator(layer):
    """Layer configuration inbound node generator.

    Args:
        layer (dict): JSON formatted layer configuration

    Yields:
        list: inbound node
    """
    for inbound_node in layer['inbound_nodes']:
        if (isinstance(inbound_node, list) and len(inbound_node) > 0 and
                isinstance(inbound_node[0], str)):
            yield [inbound_node]
        else:
            yield inbound_node


def replace_layer_name_for_connection_info(connection_info, match_name, replacement_name):
    """Updates an inbound node name.

    Args:
        connection_info (list): inbound node information
        match_name (str): inbound node name to update
        replacement_name (str): inbound node name to set

    Returns:
        list: the original inbound node if an update happened, None otherwise
    """
    # Note that is from tfmot and the connection_info structure is not really documented:
    # it is a nested list where the first item is the inbound layer name.
    # For example: [[['conv1', 0, 0, {} ]]] or [[['batch_normalization', 0, 0, {}]]]
    original_info = connection_info.copy()
    match_found = False
    if connection_info[0] == match_name:
        match_found = True
        connection_info[0] = replacement_name
    for key in connection_info[3]:
        if isinstance(connection_info[3][key], list):
            if connection_info[3][key][0] == match_name:
                match_found = True
                connection_info[3][key][0] = replacement_name
    return original_info if match_found else None


def get_layers_by_type(model, layer_type):
    """Recursively find layers matching the specified type.

    Args:
        model (keras.Model): the source model.
        layer_type (class): the Layer class to look for.

    Returns:
        list(keras.layers.Layer): a list of layers
    """
    def _get_layers(layer, layers):
        if isinstance(layer, layer_type):
            layers.append(layer)
        for attr in layer.__dict__.values():
            if isinstance(attr, Layer):
                _get_layers(attr, layers)
    layers = []
    for layer in model.layers:
        _get_layers(layer, layers)
    return layers


def update_inbound(layer_config, name, updated_inbound):
    """ Update the layer 'name' inbound in config with the provided name.

    Args:
        layer_config (dict): layer config to udpate
        name (str): name of the inbound to replace
        updated_inbound (str): new inbound name
    """
    for inbound_node in inbound_node_generator(layer_config):
        if isinstance(inbound_node, dict):
            inbound_node = inbound_node.values()
        for connection_info in inbound_node:
            replace_layer_name_for_connection_info(connection_info, name, updated_inbound)


def find_layers_pairs(model, first_type, second_type):
    """ Retrieves first_type > second_type layer pairs in a model.

    Args:
        model (keras.Model): a model
        first_type (keras.Layer): a first layer
        second_type (keras.Layer): a second layer

    Returns:
        dict: layer pairs
    """
    pairs = {}

    # Get all 'second_type' layers present in the model
    dst_layers = get_layers_by_type(model, second_type)

    # Find 'second_type' layers that have only one inbound layer that is a 'first_type' layer
    # and  that at the same time it has a single outbound ('second_type').
    for ly in dst_layers:
        parent_layer = ly.inbound_nodes[0].inbound_layers
        if isinstance(parent_layer, first_type) and len(parent_layer.outbound_nodes) == 1:
            pairs[parent_layer] = ly
    return pairs


def invert_layer_pairs(model, pairs):
    """ Edits the model configuration to invert the layer pairs and rebuilds a model.

    Args:
        model (keras.Model): a model
        pairs (dict): map between a layer and its outbound

    Returns:
        keras.Model: an updated model
    """
    # get_config documentation mentions that a copy should be made when planning to modify the
    # config
    config = deepcopy(model.get_config())
    layers = config['layers']
    new_output_name = None

    for first, second in pairs.items():
        # Retrieve layer indexes
        first_index = get_layer_index(layers, first.name)
        second_index = get_layer_index(layers, second.name)

        if isinstance(model, Sequential):
            # For Sequential model, inverting the indexes is enough
            layers[second_index], layers[first_index] = layers[first_index], layers[second_index]
        else:
            # For functional models, inbounds must be updated
            original_first_inbound = layers[first_index]['inbound_nodes'][0][0][0]

            # Update 'second' by replacing the 'first' inbound with the 'first' previous layer
            update_inbound(layers[second_index], first.name, original_first_inbound)
            # Update 'first' by replacing the original inbound with 'second'
            update_inbound(layers[first_index], original_first_inbound, second.name)

            # Then get the layers after 'second', ie. outbounds layers
            second_outbound_names = [outbound.layer.name for outbound in second.outbound_nodes]
            outbound_ids = [
                get_layer_index(layers, bn_outbound)
                for bn_outbound in second_outbound_names
            ]

            # If 'second' is the last layer (no outbound), store its name and the associated 'first'
            # name so that the model output can be updated later
            if len(outbound_ids) == 0:
                new_output_name = first.name
                last_mp = second.name

            # Finally, update the outbounds by replacing their 'second' inbound with 'first'
            for id in outbound_ids:
                update_inbound(layers[id], second.name, first.name)

    # Update the model outputs if needed
    if new_output_name:
        for index, out_layer in enumerate(config['output_layers']):
            if out_layer[0] == last_mp:
                config['output_layers'][index][0] = new_output_name

    # Reconstruct model from the config, using the cloned layers
    updated_model = model.from_config(config)

    # Restore model weights
    updated_model.set_weights(model.get_weights())
    return updated_model


def get_inbound_layers_config(layer, model_config):
    """ Retrieve the inbound layers config of 'layer'

    Args:
        layer (dict): the target layer
        model_config (dict): the config where to search the inbounds

    Returns:
        list: the config of the inbound layers
    """
    if 'inbound_nodes' in layer:
        # The config comes from a keras.Model.
        # First, recover layer names
        inbound_layer_names = []
        for inbound_node in inbound_node_generator(layer):
            if isinstance(inbound_node, dict):
                inbound_node = inbound_node.values()
            for connection_info in inbound_node:
                inbound_layer_names.append(connection_info[0])
        # Next, match name with layer config
        inbound_layers = get_layers(model_config, inbound_layer_names)
    else:
        # The config comes from a keras.Sequential.
        # Inbound layer can be found from the index of the layer
        layer_id = model_config['layers'].index(layer)
        inbound_layers = [model_config['layers'][layer_id - 1]] if layer_id > 0 else []
    return inbound_layers



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["hard_silu", "QuantizedReLU", "QuantizedActivation"]

import os
import numpy as np
import tensorflow as tf
import keras

from .layers_base import (register_quantize_target, rescale_outputs, register_aligned_inputs,
                          tensor_inputs, apply_buffer_bitwidth, QuantizedLayer)
from .quantizers import AlignedWeightQuantizer, OutputQuantizer
from .recorders import NonTrackVariable
from ..tensors import FixedPoint, QFloat, QTensor

LUT_ENV = "LUT_ENABLED"


@keras.saving.register_keras_serializable()
def hard_silu(x):
    """Hard SiLU activation function, also known as Hard Swish.

    It is defined as:

    - `0` if `if x < -3`
    - `x` if `x > 3`
    - `x * (x + 3) / 6` if `-3 <= x <= 3`

    It's a faster, piecewise linear approximation of the silu activation.

    Args:
        x(tf.Tensor): Input tensor.

    Returns:
        tf.Tensor: Output tensor.

    Reference:
        [A Howard, 2019](https://arxiv.org/abs/1905.02244)
    """
    return x * tf.nn.relu6(x + 3) / 6


@register_quantize_target(keras.layers.ReLU)
@keras.saving.register_keras_serializable()
class QuantizedReLU(QuantizedLayer):
    """Quantized version of the ReLU activation layer applicable on FixedPoint tensor.

    Args:
        max_value (float, optional): ReLU maximum value. Defaults to 6.
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """
    arg_constraints = {
        'negative_slope': 0,
        'threshold': 0}
    ignored_args = ['negative_slope', 'threshold']

    def __init__(self, *args, max_value=6, quant_config=None, **kwargs):
        super().__init__(*args, quant_config=quant_config, **kwargs)

        # Use quant_config to build quantizers
        out_quant_cfg = self.quant_config.get("output_quantizer", False)
        if out_quant_cfg:
            self.out_quantizer = OutputQuantizer(name="output_quantizer", **out_quant_cfg)
        else:
            self.out_quantizer = None
        self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=False)
        if max_value is not None:
            # Store max_value
            if isinstance(max_value, np.ndarray):
                max_value = max_value.item()
            max_value_quantizer_cfg = self.quant_config.get("max_value_quantizer", {})
            self.max_value_quantizer = AlignedWeightQuantizer(name="max_value_quantizer",
                                                              signed=False,
                                                              **max_value_quantizer_cfg)
        self.max_value = max_value

    @tensor_inputs([QTensor])
    @rescale_outputs
    def call(self, inputs):
        """ReLU activation function.

        In other terms:

            1. clip the value between 0 and :attr:`max_value`.
            2. quantize the output if an output_quantizer is set.

        Args:
            inputs (:obj:`QFloat`): the inputs tensor.

        Returns:
            :obj:`FixedPoint`: QuantizedReLU outputs.
        """
        if isinstance(inputs, FixedPoint):
            # if inputs is FixedPoint, create an equivalent QFloat with scale
            # set to 1
            inputs = QFloat(inputs, tf.constant(1.))
        # Express zero as a QFloat aligned with the inputs because this is what the
        # dispatched operations expect.
        # The actual hardware implementation will simply use a zero integer.
        zero = QFloat(FixedPoint(tf.constant(0.), inputs.fp.value_bits, inputs.fp.frac_bits),
                      inputs.scales)

        if self.max_value is None:
            # Just remove negative values
            return tf.math.maximum(inputs, zero)
        # Quantize and align max_value with the inputs
        max_value = self.max_value_quantizer(tf.cast(self.max_value, tf.float32), inputs)
        # Clip the inputs
        return tf.clip_by_value(inputs, zero, max_value)

    def get_config(self):
        config = super().get_config()
        config.update({"max_value": self.max_value})
        return config


@keras.saving.register_keras_serializable()
@register_quantize_target(keras.layers.Activation)
@register_aligned_inputs
class QuantizedActivation(QuantizedLayer, keras.layers.Activation):
    """Quantized version of `keras.layers.Activation` layer applicable on ``FixedPoint`` tensor.

    The input values are mapped through a look-up-table that simulates the activation behavior.

    Example
        >>> # Represent 2.5 as a FixedPoint
        >>> input = FixedPoint(5, value_bits=3, frac_bits=1)
        >>> # QuantizedActivation.call() maps `input` through the table to obtain
        >>> # an integer that represent the float value of tf.nn.gelu(2.5)
        >>> output = QuantizedActivation(activation="gelu")(input)
        >>> assert output.values == 80
        >>> # Or which is equivalent in float domain
        >>> max_error = 2 ** -(output.frac_bits + 1)
        >>> assert tf.abs(output.to_float() - tf.nn.gelu(2.5)) < max_error

    Args:
        activation (callable or str): Activation function. It could be a callable,
            or the name of an activation from the keras.activations namespace.
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """
    arg_constraints = {'activation': lambda: ["gelu", "swish", "Custom>hard_silu"]}
    DEFAULT_INPUT_BITWIDTH = 11
    DEFAULT_LUT_BITWIDTH = 14

    def __init__(self, activation, *args, quant_config=None, **kwargs):
        super().__init__(activation, *args, quant_config=quant_config, **kwargs)

        # Retrieve quantization parameters
        if "lut_bitwidth" not in self.quant_config:
            self.quant_config["lut_bitwidth"] = self.DEFAULT_LUT_BITWIDTH
        self.lut_bits = self.quant_config["lut_bitwidth"] - 1

        # Use quant_config to build quantizers
        out_quant_cfg = self.quant_config.get("output_quantizer", False)
        if out_quant_cfg:
            self.out_quantizer = OutputQuantizer(name="output_quantizer", **out_quant_cfg)
        else:
            self.out_quantizer = None

        # Create dynamic table and variable to save output frac_bits
        self.values = tf.lookup.experimental.MutableHashTable(key_dtype=tf.int32,
                                                              value_dtype=tf.float32,
                                                              default_value=2**self.lut_bits + 1,
                                                              name="values_table")
        self.frac_bits = NonTrackVariable("frac_bits")

    @property
    def using_lut(self):
        """Flag to specify if the inference should be given through look-up-table approach

        Returns:
            bool: True if lut is enabled, False otherwise.
        """
        value = os.environ.get(LUT_ENV, "0")
        return value == "1"

    @tf.function
    def record_values_in_table(self, value_bits, frac_bits):
        """Generate a set of inputs and outputs to record the look-up-table.

        Inputs are generated in the full range based on ``value_bits``.

        Args:
            value_bits (int): bits to define the range of values to be generated.
            frac_bits (tf.Tensor): frac_bits to convert the generated values in a FixedPoint.

        Returns:
            tf.Tensor: the expected frac_bits representing the values contained in the table.
        """
        # Generates the full range values between [-(2 ** value_bits), 2 ** value_bits - 1].
        int_max = 2 ** value_bits
        values = tf.range(-int_max, int_max, dtype=tf.int32)
        inputs = FixedPoint(values, value_bits=value_bits, frac_bits=frac_bits)

        # Forward float inputs through activation
        x = inputs.to_float()
        y = self.activation(x)

        # Apply dynamic quantization to compute output integer values.
        range_max = tf.reduce_max(tf.abs(y))
        out_frac_bits = tf.stop_gradient(FixedPoint.max_frac_bits(self.lut_bits, range_max))
        outputs = FixedPoint.quantize(y, self.lut_bits, frac_bits=out_frac_bits)

        # Insert values in table
        self.values.insert(values, outputs.values)

        # Return the static output frac_bits
        return out_frac_bits

    @tensor_inputs([FixedPoint])
    @rescale_outputs
    def call(self, inputs):
        # Values stored in the table can only be calculated if the input is per_tensor
        if not inputs.per_tensor:
            raise TypeError(f"{self.__class__.__name__} only supports per-tensor inputs.")

        self.frac_bits.init_var(inputs.frac_bits)

        # Set values in table if table is empty (low computational cost)
        if self.values.size() == 0:
            out_frac_bits = self.record_values_in_table(inputs.value_bits, inputs.frac_bits)
            out_frac_bits = tf.stop_gradient(out_frac_bits)
            self.frac_bits.set_var(out_frac_bits)

        # Look-up-table has a high cost in inference.
        # It is possible to increase the speed if we avoid it through a
        # DeQuantization-reQuantization (DQ-Q) approach:
        if not self.using_lut:
            # 1. Dequantize the inputs
            x = inputs.to_float()
            # 2. Apply activation in float domain
            x = self.activation(x)
            # 3. Requantize from static quantization approach
            outputs = FixedPoint.quantize(x, value_bits=self.lut_bits, frac_bits=self.frac_bits.var)
        else:
            # Forward inputs.values into values table
            inputs = tf.cast(inputs.values, tf.int32)
            values = self.values.lookup(inputs)

            # MutableHashTable forgets the output shape. That is why we set it explicitly.
            values.set_shape(inputs.shape)

            # Build the output FixedPoint
            outputs = FixedPoint(values, value_bits=self.lut_bits, frac_bits=self.frac_bits.var)
        return outputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_activation_ops"]

from onnx import TensorProto
from onnx.helper import make_node


def get_activation_ops(in_name, out_name, clip=False):
    """Return the activation operation chain.

    Args:
        in_name (str): the input tensor name.
        out_name (str): the required output tensor name.
        clip (bool, optional): whether to include max_value. Defaults to False.

    Returns:
        list of NodeProto: the operation chain.
    """
    nodes = [make_node("Relu", [in_name], [out_name])]
    if clip:
        nodes[0].output[0] = f"{in_name}/relu"
        # Compute bounded activation as Min(input, max_value)
        nodes += [make_node("Cast", ["max_value"], ["max_value/cast"], to=TensorProto.FLOAT),
                  make_node("Min", [nodes[0].output[0], "max_value/cast"], [out_name])]
    return nodes



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_node", "check_node_attributes", "generate_node_names"]


from .field import get_field


def get_node(nodes, op_type):
    """Helper to get a node of a specific type.

    Args:
        nodes (list of NodeProto): list of nodes.
        op_type (str): the type of the node to get.

    Returns:
        NodeProto: the node if found, None otherwise.
    """
    filtered_ops = [node for node in nodes if node.op_type == op_type]
    if len(filtered_ops) != 1:
        # Return None if not found or too many
        return None
    return filtered_ops[0]


def check_node_attributes(node, field_constraints):
    """Helper to check node attribute constraints

    Args:
        node (NodeProto): an onnx node.
        field_constraints (dict): a dictionary with the name of the field
            and the list of allowed values.
    """
    for name, values in field_constraints.items():
        try:
            current_value = get_field(node, name)
        except AssertionError:
            # When attribute is not found we continue
            continue

        if current_value not in values:
            raise ValueError(
                f"{node.name} ({node.op_type}) is supported with {name} in {values}, "
                f"not {current_value}.")


def generate_node_names(model):
    """Assigns a name to every node that does not have one.

    Args:
        model (onnx.ModelProto): the model to assign node names
    """
    for node in model.graph.node:
        if node.name == "":
            # Output tensor name is an unique name. We can use it as node name
            node.name = node.output[0]



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["QuantizedReshape"]

import tensorflow as tf
import keras

from ..layers_base import register_quantize_target, tensor_inputs, register_no_output_quantizer
from ...tensors import QTensor, FixedPoint


@register_quantize_target(keras.layers.Reshape)
@register_no_output_quantizer
@keras.saving.register_keras_serializable()
class QuantizedReshape(keras.layers.Reshape):
    """A Reshape layer that operates on quantized inputs

    Args:
        target_shape (tuple of ints): Target shape, does not include the samples
            dimension (batch size).
    """

    @tensor_inputs([QTensor])
    def call(self, inputs):
        # Return a new reshaped QTensor
        result = tf.reshape(inputs, (tf.shape(inputs)[0],) + self.target_shape)
        if not tf.executing_eagerly():
            # Set the static shape for the result since it might lost during
            # array_ops reshape, eg, some `None` dim in the result could be
            # inferred.
            if isinstance(result, FixedPoint):
                result.values.set_shape(self.compute_output_shape(inputs.shape))
            else:
                result.fp.values.set_shape(self.compute_output_shape(inputs.shape))
        return result



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from .qtensor import (
    QTensor, saturate, round_through, floor_through, pow2,
    ceil_log2, slice_tensor_by_index)
from ..debugging import assert_equal, assert_none_equal, assert_less, assert_less_equal


class FixedPoint(QTensor):
    r"""A Tensor of integer values representing fixed-point numbers

    The value_bits parameter sets the maximum integer values that can be stored:

    .. math:: int\_max = 2^{bits} - 1.

    When a FixedPoint is created, its values are clipped to [-int_max-1, int_max].

    Args:
        values (tf.Tensor): a tensor of integer values
        value_bits (int): the number of value bits.
        frac_bits (tf.Tensor): an integer tensor of fractional bits.
    """
    values: tf.Tensor = 1.0
    value_bits: int = 7
    frac_bits: tf.Tensor = 0.

    def __init__(self, values, value_bits, frac_bits):
        # We store integer values in a float tensor to speed up calculations
        if isinstance(values, tf.Tensor):
            values = tf.cast(values, tf.float32)
        else:
            values = tf.convert_to_tensor(values, dtype=tf.float32)
        # Clamp to fixed-point boundaries
        self.values = saturate(values, tf.cast(value_bits, tf.float32))
        # We store fractional bits in a float tensor to speed up calculations
        if isinstance(frac_bits, tf.Tensor):
            self.frac_bits = tf.cast(frac_bits, tf.float32)
        else:
            self.frac_bits = tf.convert_to_tensor(frac_bits, tf.float32)
        self.value_bits = value_bits

        self.shape = self.values.shape

    @property
    def name(self):
        return "no-name" if not hasattr(self.values, "name") else self.values.name

    @property
    def per_tensor(self):
        return self.frac_bits.shape.ndims == 0 or self.frac_bits.shape == (1,)

    @staticmethod
    def max_frac_bits(value_bits, ranges, clamp=True):
        """Evaluate the maximum fractional bit index for the quantization ranges.

        This method evaluates the minimum number of integer bits required to cover the specified
        quantization ranges (this can be a negative number if the ranges are strictly lower than
        0.5).

        From that it deduces the rightmost fractional bit indices.

        The resulting frac_bits can be a negative number if the ranges are higher than the biggest
        integer that can be represented with the specified value bits.

        If specified, the maximum fractional bits are clamped to the available value_bits.

        Args:
            value_bits (int): the number of value bits.
            ranges (tf:Tensor): a tensor of float quantization ranges.
            clamp (bool, optional): clamp the results to self.value_bits. Defaults to True.

        Returns:
            tf:Tensor: a tensor of fractional bits.
        """
        # Get the powers of two containing the quantization ranges
        int_bits = ceil_log2(ranges)
        # Evaluate the rightmost fractional bits (they can be negative)
        value_bits = tf.cast(value_bits, tf.float32)
        frac_bits = value_bits - int_bits
        if clamp:
            frac_bits = tf.minimum(frac_bits, value_bits)
        return frac_bits

    @staticmethod
    def quantize(x, value_bits, frac_bits=None):
        r"""Converts a float Tensor to a FixedPoint

        It converts the original float values into integer values so that:

        .. math:: {x_{int}} = round(x * 2^{frac\_bits})

        Note: :math:`2^{-frac\_bits}` represents the FixedPoint precision.

        Before returning, the resulting integer values are clipped to the
        maximum integer values that can be stored for the specified value bits:

        .. math:: [-2^{value\_bits}, 2^{value\_bits} - 1]

        If frac_bits is not specified, the method starts by evaluating the number
        of bits to dedicate to represent the integer part of the float tensor,
        clipped to value_bits:

        .. math:: int\_bits = ceil(log2(x))

        Note: this number can be negative when x < 0.5.

        It then evaluates the offset of the least significant bit of the fractional
        part of the float starting from zero. This represents the fractional bits:

        .. math:: frac\_bits = value\_bits - int\_bits

        Args:
            x (tf.Tensor): a tensor of float values.
            value_bits (int): the number of value bits
            frac_bits (tf.Tensor, optional): an integer tensor of fractional bits.
                Defaults to None.

        Returns:
            :obj:`FixedPoint`: the FixedPoint tensor
        """
        if frac_bits is None:
            if isinstance(x, int) or tf.reduce_all(x == tf.math.round(x)):
                # The input does not need to be quantized
                frac_bits = 0
            else:
                frac_bits = FixedPoint.max_frac_bits(value_bits, tf.math.reduce_max(tf.abs(x)))
        # Project float into fixed-point representation space
        x_scaled = tf.math.multiply(x, pow2(frac_bits), name="quantize")
        # Round to obtain integer values
        values = round_through(x_scaled)
        return FixedPoint(values, value_bits, frac_bits)

    @property
    def sign(self):
        """Returns the sign of the FixedPoint

        Returns:
            :obj:`FixedPoint`: the sign as a FixedPoint.
        """
        return FixedPoint(tf.math.sign(self.values), self.value_bits, 0)

    def to_float(self):
        return self.values / pow2(self.frac_bits)

    def promote(self, bits):
        """Increase the number of value bits

        Args:
            bits (int): the new number of value bits

        Returns:
            :obj:`FixedPoint`: a FixedPoint with increased value bits
        """
        if not isinstance(bits, int):
            raise TypeError("Bitwidth must be an integer")
        if bits < 0:
            raise ValueError("Invalid bitwidth")
        if bits < self.value_bits:
            raise ValueError(f"Cannot reduce value bits from {self.value_bits} to {bits}: "
                             "use a quantizer instead")
        return FixedPoint(self.values, bits, self.frac_bits)

    def align(self, other, value_bits=None):
        """Align fractional bits

        This returns an equivalent FixedPoint with a scalar fractional bit
        corresponding to the maximum of the current and other FixedPoint on all
        channels.

        This is required before performing an operation that adds or subtracts
        elements along the last dimension, to make sure all these elements are
        in the same scale.

        Args:
            other (:obj:`FixedPoint`): a FixedPoint to align to
            value_bits (int, optional): the target value bits. Defaults to None.

        Returns:
            tuple(:obj:`FixedPoint`, tf.Tensor): a new FixedPoint with aligned
            fractional bits and the shift that was applied.
        """
        if not isinstance(other, FixedPoint):
            raise ValueError("Other must be a FixedPoint.")
        max_frac_bits = tf.math.maximum(self.frac_bits, other.frac_bits)
        return self.upscale(max_frac_bits, value_bits)

    def downscale(self, frac_bits):
        """Reduce the precision of a FixedPoint

        Args:
            frac_bits (tf.Tensor): the target fractional bits

        Returns:
            :obj:`FixedPoint`: the downscaled FixedPoint
        """
        frac_bits = tf.cast(frac_bits, tf.float32)
        assert_less_equal(frac_bits, self.frac_bits,
                          f"Cannot reduce {self.name} precision because the target precision "
                          f"({frac_bits}) is higher than the current one ({self.frac_bits})")
        return self.rescale(frac_bits)

    def rescale(self, frac_bits, value_bits=None):
        """Rescale a FixedPoint to a specified precision and bitwidth

        This primarily rescales the FixedPoint values to match the precision
        specified by the target fractional bits.

        Optionally, this adjusts the value bits to the specified bitwidth.

        The rescaling operation is:

        - a left shift of the values when their precision increases,
        - a rounded right shift of the values when their precision decreases.

        This method can be used to:

        - compress a FixedPoint to a lower bitwidth after having reduced its precision,
        - expand a FixedPoint to a larger bitwidth after having increased its precision.

        Args:
            frac_bits (tf.Tensor): the target fractional bits
            value_bits (int, optional): the target value bits

        Returns:
            :obj:`FixedPoint`: the rescaled FixedPoint
        """
        if value_bits is None:
            value_bits = self.value_bits
        frac_bits = tf.cast(frac_bits, tf.float32)
        # Evaluate the shift to apply to reach the target precision
        shift = frac_bits - self.frac_bits
        # The shift can be positive (left-shift) or negative (rounded right-shift)
        # A positive shift exceeding the target bitwidth always leads to a saturation
        assert_less(shift, value_bits,
                    f"Cannot rescale {self.name} to {frac_bits} fractional bits as it will saturate"
                    f" its {value_bits} buffer")
        # The integer operations are simulated in float arithmetics:
        # - the shifts by multiplying by the positive/negative shift power of two,
        # - the rounding by adding 0.5 and flooring (to workaround banker's rounding).
        # For positive shift, the results being integer, the rounding has no effect.
        # We therefore apply the same operations for both shifts.
        values = tf.math.multiply(self.values, pow2(shift), name="rescale")
        values = floor_through(values + 0.5)
        # return a new FixedPoint with the target precision and bitwidth
        return FixedPoint(values, value_bits, frac_bits), shift

    def upscale(self, frac_bits, value_bits=None):
        """Align a FixedPoint to a specified precision

        The target precision must be higher than the current one.

        Args:
            frac_bits (tf.Tensor): the target fractional bits
            value_bits (int, optional): the target value bits

        Returns:
            :obj:`FixedPoint`: the upscaled FixedPoint
        """
        frac_bits = tf.cast(frac_bits, tf.float32)
        assert_less_equal(self.frac_bits, frac_bits,
                          f"Cannot increase {self.name} precision because the target precision "
                          f"({frac_bits}) is lower than the current one ({self.frac_bits})")
        return self.rescale(frac_bits, value_bits)

    def expand(self, value_bits):
        """Expand the FixedPoint to the specified bitwidth

        This returns an equivalent FixedPoint with a higher or equal number of
        value bits and a scalar fractional bit corresponding to the maximum of
        the initial fractional bits on all channels.

        This is mostly used to recover a per-tensor FixedPoint that has been
        compressed to a lower number of value bits.

        Args:
            value_bits (int): the target value_bits

        Returns:
            tuple(:obj:`FixedPoint`, tf.Tensor): a new FixedPoint with expanded
            fractional bits and the shift that was applied.
        """
        if value_bits < self.value_bits:
            raise ValueError(
                f"Cannot reduce {self.name} bitwidth from {self.value_bits} to {value_bits}:"
                " use a quantizer instead.")
        max_frac_bits = tf.reduce_max(self.frac_bits)
        return self.upscale(max_frac_bits, value_bits)

    @staticmethod
    def _rshift(values, shift):
        return floor_through(values / pow2(shift))

    @staticmethod
    def _lshift(values, shift):
        return tf.math.multiply(values, pow2(shift), name="lshift")

    def shift(self, s):
        """Apply a tensor-wide left or right shift.

        This takes a tensor of shift values and apply them on each item of the
        FixedPoint values.

        The shift values should positive or negative integer:

        - if the value is positive, it is a left-shift,
        - if the value is negative, it is a right-shift.

        The resulting FixedPoint has the same value bits and fractional bits as
        the source FixedPoint, which means that clipping is applied on
        left-shift and flooring is applied on right-shift.

        Args:
            s (tf.Tensor): the shift values for each pixel.

        Returns:
            :obj:`FixedPoint`: the result as a FixedPoint
        """
        values = tf.math.multiply(self.values, pow2(s), name="shift")
        values = floor_through(values)
        return FixedPoint(values, self.value_bits, self.frac_bits)

    def __rshift__(self, shift):
        """Right shift the FixedPoint values

        This operation has no direct equivalent in float arithmetics: it corresponds to a division
        of the corresponding float by a power-of-two, then a flooring to the quantization interval.

        Args:
            shift (tf.Tensor): the power of 2 to divide by

        Returns:
            :obj:`FixedPoint`: the result as a FixedPoint
        """
        shift = tf.cast(shift, tf.float32)
        assert_less_equal(0, shift, "Shift must be all positive")
        assert_equal(tf.rank(shift) <= tf.rank(self.frac_bits), True,
                     "The shift's rank must be less than or equal to the rank of frac_bits. "
                     f"Received {tf.rank(shift)} > {tf.rank(self.frac_bits)}."
                     "That means it is not possible to fold the shift into the FixedPoint. "
                     "Please use FixedPoint.shift instead of '>>'.")

        # The shift can be folded into the fractional bits,
        s_frac_bits = self.frac_bits + shift
        # keeping the same values
        s_values = self.values
        # Return a new FixedPoint with updated fractional bits,
        # which is equivalent in hardward without performing any operation
        return FixedPoint(s_values, self.value_bits, s_frac_bits)

    def __lshift__(self, shift):
        """Left shift the FixedPoint values

        This operation has no direct equivalent in float arithmetics: it corresponds to a
        multiplication of the corresponding float by a power-of-two, then a flooring to the
        quantization interval.

        Args:
            shift (tf.Tensor): the power of 2 to multiply by

        Returns:
            :obj:`FixedPoint`: the result as a FixedPoint
        """
        assert_less_equal(0, shift, "Shift must be all positive")
        # Simply apply the shift on the values
        s_values = FixedPoint._lshift(self.values, shift)
        # Return a new FixedPoint with updated values
        return FixedPoint(s_values, self.value_bits, self.frac_bits)

    def _align_values(self, other):
        # The sub fractional bits are the max of both terms
        frac_bits = tf.math.maximum(self.frac_bits, other.frac_bits)
        self_values = FixedPoint._lshift(
            self.values, (frac_bits - self.frac_bits))
        other_values = FixedPoint._lshift(
            other.values, (frac_bits - other.frac_bits))
        return frac_bits, self_values, other_values

    def __add__(self, other):
        if isinstance(other, int):
            # Convert integer into a 32-bit fixed-point with no fractional bits,
            # aligned with the current FixedPoint
            return self + FixedPoint.quantize(other, 32, self.frac_bits)
        elif isinstance(other, FixedPoint):
            # Check that self and other are aligned
            assert_equal(self.frac_bits, other.frac_bits,
                         message=f"{self.name} and {other.name} are not aligned")
            # Return a new FixedPoint
            return FixedPoint(self.values + other.values, self.value_bits, self.frac_bits)
        raise TypeError(
            f"Unsupported operand type(s) for +: 'FixedPoint' and '{type(other)}'")

    def __sub__(self, other):
        if isinstance(other, int):
            # Convert integer into a 32-bit fixed-point with no fractional bits,
            # aligned with the current FixedPoint
            return self - FixedPoint.quantize(other, 32, self.frac_bits)
        elif isinstance(other, FixedPoint):
            # Check that self and other are aligned
            assert_equal(self.frac_bits, other.frac_bits,
                         message=f"{self.name} and {other.name} are not aligned")
            # Return a new FixedPoint
            return FixedPoint(self.values - other.values, self.value_bits, self.frac_bits)
        raise TypeError(
            f"Unsupported operand type(s) for -: 'FixedPoint' and '{type(other)}'")

    def __truediv__(self, other):
        @tf.function
        @tf.custom_gradient
        def truncate(x):
            # Remove decimal part from x, This is to obtain a value that
            # matches the result estimated in C when working with ints.
            rounded = tf.cast(tf.cast(x, tf.int32), tf.float32)

            def grad(upstream):
                return upstream
            return rounded, grad

        if isinstance(other, int):
            return self / FixedPoint(other, 32, 0)
        elif isinstance(other, FixedPoint):
            assert_none_equal(other.values, 0.0, "Cannot divide by 0.")
            # The division between fixed-point is straight-forward
            d_values = truncate(self.values / other.values)
            # Return a new FixedPoint whose frac bits from other is subtracted.
            return FixedPoint(d_values, self.value_bits, self.frac_bits - other.frac_bits)
        raise TypeError(
            f"Unsupported operand type(s) for /: 'FixedPoint' and '{type(other)}'")

    def __getitem__(self, idx):

        r"""Retrieve a slice or element from the FixedPoint tensor.

        This function allows for slicing or indexing the FixedPoint tensor in a similar way
        to a standard TensorFlow tensor. It handles the slicing of both the values
        and the frac_bits of the FixedPoint tensor, ensuring that the tensor
        representation remains consistent.

        Args:
            idx (int, slice, or tuple): The index or slice to retrieve. It can be an integer,
                a slice, or a tuple of integers/slices. The tuple can
                also include the Ellipsis (`...`) to denote a slice
                over remaining dimensions.

        Returns:
            FixedPoint: A new FixedPoint instance containing the sliced values and corresponding
                frac_bits.
        """

        sliced_values = self.values[idx]

        if self.per_tensor:
            return FixedPoint(sliced_values, self.value_bits, self.frac_bits)

        else:
            values_rank = len(self.shape)
            sliced_frac_bits = slice_tensor_by_index(idx, values_rank, self.frac_bits)
            return FixedPoint(sliced_values, self.value_bits, sliced_frac_bits)

    def __pow__(self, power):
        if isinstance(power, int):
            if power == 0:
                return FixedPoint(tf.ones_like(self.values), self.value_bits, 0)
            elif power == 1:
                return FixedPoint(self.values, self.value_bits, self.frac_bits)
            elif power > 1:
                return self * self ** (power - 1)
            else:
                raise NotImplementedError(
                    "Negative powers are not implemented yet")
        raise TypeError(
            f"Unsupported operand type(s) for **: 'FixedPoint' and '{type(power)}'")

    def __gt__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for >: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values > o_values

    def __ge__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for >=: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values >= o_values

    def __eq__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for ==: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values == o_values

    def __ne__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"unsupported operand type(s) for !=: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values != o_values

    def __lt__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for <: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values < o_values

    def __le__(self, other):
        if not isinstance(other, FixedPoint):
            raise TypeError(
                f"Unsupported operand type(s) for <=: 'FixedPoint' and '{type(other)}'")
        _, s_values, o_values = self._align_values(other)
        return s_values <= o_values

    def abs(self):
        """Returns the absolute value of the FixedPoint

        Returns:
            :obj:`FixedPoint`: the absolute value.
        """
        return FixedPoint(tf.math.abs(self.values), self.value_bits, self.frac_bits)

    def floor(self):
        """Floors the FixedPoint

        Returns:
            tuple(:obj:`FixedPoint`, tf.Tensor): a new FixedPoint without
            fractional bits and the shift that was applied.
        """
        # Divide values to remove fractional bits
        values = FixedPoint._rshift(self.values, self.frac_bits)
        return FixedPoint(values, self.value_bits, 0), self.frac_bits



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["QuantizedAdd", "get_qadd", "QuantizedConcat", "get_qconcat"]

import numpy as np

from onnx import TensorProto as TP
from onnx.helper import make_node

from .base_layer import OnnxLayer, register_node_format
from .subgraph_ops import cast_tensors_to, get_scale_out_ops, get_input_shift_ops
from .subgraph_ops.activation import get_activation_ops
from ..graph_tools import TENSOR_SHAPE, get_tensor_shape, get_node
from ..quantization.weights import align_to
from ..quantization.outputs import downscale_fp
from ..graph_tools import get_field


def get_qadd(nodes, graph):
    # Both inputs should not be constants
    weight_names = [x.name for x in graph.initializer]
    if nodes[0].input[0] in weight_names or nodes[0].input[1] in weight_names:
        raise ValueError("Unsupported Add: inputs should be tensors.")

    add_node = nodes[0]
    add_name = add_node.name

    act_node = get_node(nodes, 'Relu')
    qadd = QuantizedAdd(name=add_name, activation=bool(act_node))

    return qadd


def get_qconcat(nodes, graph):
    concat_node = nodes[0]
    concat_name = concat_node.name
    concat_axis = get_field(concat_node, "axis")
    return QuantizedConcat(name=concat_name, axis=concat_axis)


@register_node_format(requires_fp_inputs=True)
class QuantizedAdd(OnnxLayer):
    """Intermediate representation of Add() as an exportable node.

    Args:
        activation (bool, optional): whether to apply relu operation. Defaults to False.
        name (str, optional): the node name. Defaults to ''.
    """

    def __init__(self, activation=False, name=''):
        super().__init__("QuantizedAdd", name=name)

        # Save properties need to serialize operation name
        self.serialize_attr["activation"] = activation

    @property
    def op_type(self):
        # Overriding op_type as ReLU comes after Scaling
        op_name = self.base_name
        if self.serialize_attr["scale"]:
            op_name += "Scaled"
        if self.serialize_attr["activation"]:
            op_name += "ReLU"
        return op_name

    def __build__(self, a_input_ts, b_input_ts, downscale=True):
        assert a_input_ts.dtype == b_input_ts.dtype == np.int8
        assert a_input_ts.shape == b_input_ts.shape

        # The chain of operations is modified if downscale is needed
        self.serialize_attr["scale"] = downscale

        # Compute output shape
        output_type = "int8" if downscale else "int32"
        output_ts = TENSOR_SHAPE(a_input_ts.shape, np.dtype(output_type))
        return output_ts

    def __quantize__(self, a_qinput, b_qinput, out_tensor_range, force_fp=False):
        def _round_pot(x):
            return 2.0**np.round(np.log2(x))

        x1_scale = a_qinput.weights["scale"]
        x2_scale = b_qinput.weights["scale"]
        # This quantization is feasible if and only if input scales are power-of-two
        np.testing.assert_array_equal(x1_scale, _round_pot(x1_scale), "Required a power-of-two")
        np.testing.assert_array_equal(x2_scale, _round_pot(x2_scale), "Required a power-of-two")

        # Prepare tensors list with unique names
        prefix = self.name + "_"

        # Transpose scales to align with channels
        output_shape = get_tensor_shape(self.output)
        x1_scale = align_to(x1_scale, len(output_shape))
        x2_scale = align_to(x2_scale, len(output_shape))

        # We expected input scales are a power-of-two. Take i_scale as a max of both scales
        i_scale = np.maximum(x1_scale, x2_scale)

        # Shift to apply for each input will be
        weights_dict = {prefix + "x1_shift": (i_scale / x1_scale).astype("int32"),
                        prefix + "x2_shift": (i_scale / x2_scale).astype("int32")}

        if "Scaled" not in self.op_type:
            out_scale = i_scale.squeeze()
        else:
            # Now consider calibrated output range
            scale, s_out, out_scale = downscale_fp(out_tensor_range, i_scale, bitwidth=8)
            # Add does not have output scale. We fold scale into shift as a power-of-two.
            # This will force an 'output scale' = 1
            s_out = 2.0**(np.log2(s_out) - np.ceil(np.log2(scale)))
            weights_dict.update({prefix + "S_out": s_out.astype(np.float32)})

        # Return quantized weights and output scale
        return weights_dict, out_scale

    @staticmethod
    def build_subgraph(op_type):
        # Cast inputs and shift to float.
        nodes, t_names = cast_tensors_to(["X", "Y", "Xs", "Ys"])

        # Align inputs with input shift
        nodes += get_input_shift_ops(t_names[0], t_names[2], "Xshifted")
        nodes += get_input_shift_ops(t_names[1], t_names[3], "Yshifted")

        # Perform addition
        nodes.append(make_node("Add", inputs=["Xshifted", "Yshifted"], outputs=["Zi"]))

        # Apply final output shift (optional)
        if "Scaled" in op_type:
            shift_nodes, shift_t_names = cast_tensors_to(["", "Shift"])
            nodes += shift_nodes
            nodes += get_scale_out_ops("Zi", "Zscaled", *shift_t_names)
            nodes.append(make_node("Cast", ["Zscaled"], ["Z"], to=TP.INT8))
        else:
            nodes.append(make_node("Cast", ["Zi"], ["Z"], to=TP.INT32))

        # Activation (optional)
        if "ReLU" in op_type:
            # rename previous output
            nodes[-1].output[0] = "Z_no_act"
            nodes += get_activation_ops(nodes[-1].output[0], "Z")

        return nodes


class QuantizedConcat(OnnxLayer):
    """Intermediate representation of Concatenate() as an exportable node.

    Args:
        name (str, optional): the node name. Defaults to ''.
    """

    def __init__(self, axis,  name=''):
        if axis != 1:
            raise NotImplementedError("Axis != 1 in Concat is not supported yet.")
        super().__init__("QuantizedConcat", axis=axis, name=name)

    def __build__(self, a_input_ts, b_input_ts, downscale=False):
        # return the concatenated shape
        b, c, w, h = a_input_ts.shape
        concat_output_shape = (b, a_input_ts.shape[1] + b_input_ts.shape[1], w, h)
        output_ts = TENSOR_SHAPE(concat_output_shape, a_input_ts.dtype)
        return output_ts

    def __quantize__(self, a_qinput, b_qinput, out_tensor_range, force_fp=False):
        # get scale and broadcast to input shape in case of per tensor
        x1_scale = a_qinput.weights["scale"]
        x1_shape = get_tensor_shape(a_qinput.output)
        x1_scale = x1_scale * np.ones(x1_shape[1])
        x2_scale = b_qinput.weights["scale"]
        x2_shape = get_tensor_shape(b_qinput.output)
        x2_scale = x2_scale * np.ones(x2_shape[1])
        # concatenate scales
        out_scale = np.concatenate((x1_scale, x2_scale))
        return {}, out_scale

    @staticmethod
    def build_subgraph(op_type):
        # Perform concatenation
        return [make_node("Concat", inputs=["X", "Y"], outputs=["Zi"], axis=1)]



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Common utility methods used in quantization models.
"""

__all__ = ['apply_weights_to_model', 'get_model_input_dtype']

import warnings
import tensorflow as tf

from ..layers import StatefulRecurrent


def apply_weights_to_model(model, weights, verbose=True):
    """Loads weights from a dictionary and apply it to a model.

    Go through the dictionary of weights, find the corresponding variable in the
    model and partially load its weights.

    Args:
        model (keras.Model): the model to update
        weights (dict): the dictionary of weights
        verbose (bool, optional): if True, throw warning messages if a dict item is not found in the
            model. Defaults to True.
    """
    if len(weights) == 0:
        warnings.warn("There is no weight to apply to the model.")
        return

    # Go through the dictionary of weights with each item
    for key, value in weights.items():
        value_applied = False
        for dest_var in model.variables:
            if key == dest_var.name:
                # Apply the current item value
                dest_var.assign(value)
                value_applied = True
                break
        if not value_applied and verbose:
            warnings.warn(f"Variable '{key}' not found in the model.")


def get_model_input_dtype(model):
    """Retrieve the common input dtype for a model

    Handle image like samples (channels in [1, 3]) as uint8, recurrent TENNs as int16 and
    other as int8

    Args:
        model (keras.Model or keras.Sequential): the model to get the input dtype.

    Returns:
        tf.dtype: the expected input type for the model
    """
    if any(isinstance(ly, StatefulRecurrent) for ly in model.layers):
        return tf.int16
    return tf.uint8 if model.input_shape[-1] in [1, 3] else tf.int8



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["extract_submodel_from_qnode"]

import onnx

from ...onnx_support.quantization import ONNXModel


def extract_submodel_from_qnode(qnode, qmodel, model_to_extract):
    """Extract a submodel of 'model_to_extract' that represents the 'qnode' chain of operations

    Args:
        qnode (onnx.NodeProto): the reference quantized node
        qmodel (ONNXModel): the model containing qnode
        model_to_extract (ONNXModel): Model to extract the chain of operations

    Returns:
        ONNXModel: a submodel of 'model_to_extract'
    """
    def _get_input_tensor_name(node, idx, model, skip_ops=[]):
        parent = model.get_parent(node, idx)
        if parent is not None and parent.op_type in skip_ops:
            return _get_input_tensor_name(parent, 0, model, skip_ops=skip_ops)
        return node.input[idx]

    if qnode.domain != "com.brainchip":
        raise ValueError(f"Unrecognized '{qnode.op_type}' quantized node.")
    elif qnode.op_type in ["InputQuantizer", "Dequantizer"]:
        raise ValueError(f"It is not possible to extract a submodel from {qnode.op_type}, "
                         f"since it does not represent any node in {model_to_extract.name}.")

    # Search node inputs
    input_names = []
    initializer_names = qmodel.get_initializer_name_set()
    for idx, iname in enumerate(qnode.input):
        if iname in initializer_names:
            # This input is an initializer (weight). Skip it
            continue

        if qnode not in model_to_extract.nodes():
            # iname will be replaced by input graph name if node is linked to an InputQuantizer,
            # since first float node is quantized as InputQuantizer + its quantized node
            # (e.g. Conv -> [InputQuantizer, QuantizedConv])
            iname = _get_input_tensor_name(qnode, idx, qmodel, skip_ops=["InputQuantizer"])
            # Nothing to do if node exist in 'model_to_extract'
        input_names.append(iname)

    # Extract submodel from 'model_to_extract_nodes' with nodes between the tensor names:
    # input_names -> node_1 -> node_2 -> ... -> node_n -> qnode.output[0]
    output_names = qnode.output[:1]
    try:
        extractor = onnx.utils.Extractor(model_to_extract.model)
        submodel = extractor.extract_model(input_names, output_names)
    except Exception as e:
        raise RuntimeError(f"{model_to_extract.name} does not contain tensor with names "
                           f"{input_names + output_names}. More detail in:\n{str(e)}")
    return ONNXModel(submodel)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["QuantizedDropout"]

import keras

from .layers_base import (register_quantize_target, tensor_inputs, register_no_output_quantizer,
                          QuantizedLayer)
from ..tensors import QTensor


@register_quantize_target(keras.layers.Dropout)
@register_no_output_quantizer
@keras.saving.register_keras_serializable()
class QuantizedDropout(QuantizedLayer, keras.layers.Dropout):
    """ A dropout layer that operates on quantized inputs and weights.

    It is only implemented as a passthrough.
    """

    @tensor_inputs([QTensor])
    def call(self, inputs):
        # QuantizedDropout act as a pass through.
        return inputs

    def get_config(self):
        return keras.layers.Dropout.get_config(self)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["array_to_tp", "infer_partial_io", "get_tensor_shape", "get_tensor_dtype",
           "value_info_to_tensor_shape", "TENSOR_SHAPE", "find_value_info_by_name"]

import numpy as np
from collections import namedtuple

import onnx
import onnx.numpy_helper

TENSOR_SHAPE = namedtuple('TensorShape', ['shape', 'dtype'])


def get_tensor_shape(tensor_value_info):
    """Helper to extract the shape in a ValueInfoProto.

    Args:
        tensor_value_info (ValueInfoProto): the value info to read the shape.

    Returns:
        tuple: the tensor shape
    """
    assert isinstance(tensor_value_info, onnx.ValueInfoProto)
    tshape = tensor_value_info.type.tensor_type.shape.dim
    tshape = tuple(el.dim_param or None if el.dim_value == 0 else el.dim_value for el in tshape)
    if len(tshape) == 0 or not all(isinstance(dim, int) and dim != 0 for dim in tshape[1:]):
        raise RuntimeError(f"{tensor_value_info.name} shape ({tshape}) must be static "
                           "and it should has at least one known-dimension.")
    return tshape


def get_tensor_dtype(tensor_value_info):
    """Helper to extract the np.dtype in a ValueInfoProto.

    Args:
        tensor_value_info (ValueInfoProto): the value info to read the dtype.

    Returns:
        np.dtype: the tensor dtype
    """
    assert isinstance(tensor_value_info, onnx.ValueInfoProto)
    tensor_type = tensor_value_info.type.tensor_type.elem_type
    return onnx.helper.tensor_dtype_to_np_dtype(tensor_type)


def value_info_to_tensor_shape(x):
    """Helper to extract the shape and dtype contains in the input.

    Args:
        x (ValueInfoProto): the value info to read the shape.

    Returns:
        tuple: the tensor shape and dtype
    """
    return TENSOR_SHAPE(get_tensor_shape(x), get_tensor_dtype(x))


def array_to_tp(**kwargs):
    """Transform a numpy array list to TensorProto list

    Args:
        kwargs (dict, optional): a list of numpy arrays. Defaults to {}.

    Returns:
        list of TensorProto: the list of tensor proto.
    """
    # Transform each input in a TensorProto
    tensors = []
    for name, x in kwargs.items():
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        tensors.append(onnx.numpy_helper.from_array(x, name))
    return tensors


def infer_partial_io(nodes, exclude=[]):
    """Infer the partial inputs/outputs for a list of 'connected' nodes.

    Args:
        nodes (list of NodeProto): the nodes list.
        exclude (list of str): exclude tensors with these names. Defaults to [].

    Returns:
        list, list: the inputs outputs infered.
    """
    # Search partial outputs
    def _extract_unique_not_null_elems(elems, exclude=[]):
        return sorted(set(el for el in elems if el not in exclude and el), key=elems.index)

    # Infer ordered, not null and unique input/output names
    all_inputs = sum([list(node.input) for node in nodes], [])
    all_outputs = sum([list(node.output) for node in nodes], [])
    inputs = _extract_unique_not_null_elems(all_inputs, exclude=all_outputs + exclude)
    outputs = _extract_unique_not_null_elems(all_outputs, exclude=all_inputs + exclude)
    return inputs, outputs


def find_value_info_by_name(graph, tensor_name):
    """Return a value info by its name

    Args:
        graph (GraphProto): the onnx graph
        tensor_name (str): the tensor name

    Returns:
        ValueInfoProto: the value info
    """
    for value_info in [*graph.input, *graph.output, *graph.value_info]:
        if value_info.name == tensor_name:
            return value_info



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["WeightQuantizer"]

import tensorflow as tf
import keras

from ...tensors import QTensor, QFloat, FixedPoint
from ..recorders import QFloatRecorder, FixedPointRecorder
from .quantizers import Quantizer


@keras.saving.register_keras_serializable()
class WeightQuantizer(Quantizer):
    """A uniform quantizer that converts a float Tensor to a QFloat representation.

    In order, the WeightQuantizer:

        - evaluates the scales required to align the values on optimal ranges for FixedPoint
          quantization,
        - quantizes the rescaled Tensor as a FixedPoint and returns a QFloat.

    Args:
        bitwidth (int, optional): the quantization bitwidth, defaults to 4.
        signed (bool, optional): whether the quantizer expects signed values or unsigned.
            Defaults to True.
        axis (int, optional): the quantization range is a scalar (None) or a vector corresponding to
            the given axis. Defaults to -1.
        fp_quantizer (bool, optional): True to enable FixedPoint quantization, QFloat otherwise.
            Defaults to False.
    """

    def __init__(self, bitwidth=4, signed=True, axis=-1, fp_quantizer=False, **kwargs):
        super().__init__(bitwidth, signed, **kwargs)
        self.axis = axis
        self.fp_quantizer = fp_quantizer
        qw_name = self.name + "/qweights"
        if self.fp_quantizer:
            self.qweights = FixedPointRecorder(qw_name)
        else:
            self.qweights = QFloatRecorder(qw_name)

    def build(self, input_shape):
        """Build the layer.

        Args:
            input_shape (list): the shape of input tensor.
        """
        super().build(input_shape)
        # Convert axis to a list of int
        if self.axis is not None and len(input_shape) > 1:
            axis_list = list(range(len(input_shape)))
            positive_axis = self.axis if self.axis >= 0 else len(axis_list) + self.axis
            if positive_axis < 0 or positive_axis > len(axis_list) - 1:
                raise ValueError(f"Axis {self.axis} is not valid in {self.name}.")
            axis_list.pop(positive_axis)
            self._axis = axis_list
        else:
            self._axis = None

    def call(self, inputs):
        """Quantize the float inputs

        The quantization is done in two steps:

            1. Compute the quantization ranges,
            2. Quantize the inputs.

        Args:
            inputs (tf.Tensor): the inputs tensor.

        Returns:
            :obj:`QFloat`: the quantized tensor.
        """
        if isinstance(inputs, QTensor):
            raise ValueError(
                f"{type(inputs)} input is not supported. WeightQuantizer only accepts float"
                " inputs.")

        # Compute the quantization ranges from the inputs
        ranges = tf.math.reduce_max(tf.math.abs(inputs), self._axis)
        if self.axis in [-2, 2] and inputs.shape[-1] == 1 and len(inputs.shape) > 2:
            # Expand the shape of the ranges so that it is broacastable on the inputs
            ranges = tf.expand_dims(ranges, -1)

        if self.fp_quantizer:
            # Build a FixedPoint when scale is disabled
            frac_bits = tf.stop_gradient(FixedPoint.max_frac_bits(self.value_bits, ranges))
            qweights = FixedPoint.quantize(inputs, self.value_bits, frac_bits)
        else:
            # Evaluate the scales to align on the optimal quantization ranges
            scales = QFloat.optimal_scales(ranges, self.value_bits)
            # Clip scales lower bound to avoid quantizing very tiny values. Minimum is defined
            # as 1e-6, value low enough to be considered as zero.
            scales = tf.maximum(scales, 1e-6)
            # Since we use the optimal quantization ranges [-int_max -1, int_max], the inner
            # FixedPoint can be quantized with exactly zero fractional bits
            qweights = QFloat.quantize(inputs, self.value_bits, scales, 0.)
        # Record the quantized weights (it does nothing if recording is disabled)
        self.qweights(qweights)
        return qweights

    def get_config(self):
        """Get the config of the layer.

        Returns:
            dict: the config of the layer.
        """
        config = super().get_config()
        config.update({"axis": self.axis, "fp_quantizer": self.fp_quantizer})
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
QuantizedDepthwiseConv2D layer definition.
"""

__all__ = ["QuantizedDepthwiseConv2D", "DepthwiseConv2DTranspose",
           "QuantizedDepthwiseConv2DTranspose"]


import tensorflow as tf
import keras

from keras import backend
from keras.layers import DepthwiseConv2D

from .layers_base import (register_quantize_target, rescale_outputs, tensor_inputs,
                          neural_layer_init, register_aligned_inputs, check_arg_constraints,
                          QuantizedLayer)
from .convolution import deconv_output_length
from ..tensors import FixedPoint, QFloat


@register_quantize_target(DepthwiseConv2D)
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedDepthwiseConv2D(QuantizedLayer, DepthwiseConv2D):
    """ A depthwise convolutional layer that operates on quantized inputs and weights.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """
    @neural_layer_init(False)
    def __init__(self, *args, quant_config=None, **kwargs):
        # Override WeightQuantizer axis to -2 which corresponds to the channel dimension of the
        # depthwise operation.
        self.weight_quantizer.axis = -2
        self.quant_config['weight_quantizer']['axis'] = -2

        # Limit buffer bitwidth to 27
        self.quant_config['buffer_bitwidth'] = min(28, self.quant_config['buffer_bitwidth'])
        self.buffer_bitwidth = self.quant_config['buffer_bitwidth'] - 1

    @tensor_inputs([FixedPoint, tf.Tensor])
    @rescale_outputs
    def call(self, inputs):
        # Quantize the weights
        depthwise_kernel = self.weight_quantizer(self.depthwise_kernel)

        outputs = backend.depthwise_conv2d(
            inputs,
            depthwise_kernel,
            strides=self.strides,
            padding=self.padding,
            dilation_rate=self.dilation_rate,
            data_format=self.data_format)

        if self.use_bias:
            # Quantize bias and align it on the outputs
            bias = self.bias_quantizer(self.bias, outputs)
            outputs = tf.add(outputs, bias)

        return outputs


@keras.saving.register_keras_serializable()
class DepthwiseConv2DTranspose(DepthwiseConv2D):
    """ A transposed depthwise convolutional layer.

    It performs a transposed depthwise convolution on inputs.
    """
    arg_constraints = {
        'dilation_rate': lambda: [1, [1, 1], (1, 1)],
        'depth_multiplier': 1,
        'strides': lambda: [2, [2, 2], (2, 2)],
        'padding': 'same'
    }

    def __init__(self, *args, **kwargs):
        check_arg_constraints(self, kwargs)
        super().__init__(*args, **kwargs)

    def build(self, input_shape):
        # Ensure variables are build with the appropriate name
        with tf.name_scope(self.name + '/'):
            super().build(input_shape)

    def call(self, inputs):
        # Infer the dynamic output shape
        inputs_shape = tf.shape(inputs)
        out_height = deconv_output_length(
            inputs_shape[1],
            self.kernel_size[0],
            padding=self.padding,
            stride=self.strides[0],
            dilation=self.dilation_rate[0])
        out_width = deconv_output_length(
            inputs_shape[2],
            self.kernel_size[1],
            padding=self.padding,
            stride=self.strides[1],
            dilation=self.dilation_rate[1])
        output_shape = tf.stack((inputs_shape[0], out_height, out_width, 1))
        # Duplicate output_shape to create a placeholder that could be iterated in
        # tf.vectorized_map, making keras happy.
        output_shape = tf.repeat([output_shape], repeats=inputs_shape[-1], axis=0)

        # Inputs and kernels must be transposed to have their channel
        # dimension first because the tf.vectorized_map call that follows will
        # unpack them on dimension 0. The channel dimension is virtually
        # restored using expand_dims so that elements have the appropriate
        # shape for the conv2d_transpose call (with a channel dimension of 1
        # which is expected in the depthwise process).
        inputs_channel_first = tf.transpose(inputs, (3, 0, 1, 2))
        inputs_channel_first = tf.expand_dims(inputs_channel_first, -1)
        kernel_channel_first = tf.transpose(
            self.depthwise_kernel, (2, 0, 1, 3))
        kernel_channel_first = tf.expand_dims(kernel_channel_first, -2)

        dw_outputs = tf.vectorized_map(
            lambda x: backend.conv2d_transpose(x[0],
                                               x[1],
                                               output_shape=x[2],
                                               strides=self.strides,
                                               padding=self.padding),
            (inputs_channel_first, kernel_channel_first, output_shape))
        outputs = tf.transpose(tf.squeeze(dw_outputs, axis=-1), (1, 2, 3, 0))

        # Last dimension is lost when building layer outputs in model.
        outputs = tf.reshape(outputs, (inputs_shape[0], out_height, out_width, inputs_shape[-1]))

        if self.use_bias:
            outputs = tf.add(outputs, self.bias)
        return outputs


@register_quantize_target(DepthwiseConv2DTranspose)
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedDepthwiseConv2DTranspose(QuantizedLayer, DepthwiseConv2DTranspose):
    """ A transposed depthwise convolutional layer that operates on quantized
    inputs and weights.

    Args:
        quant_config (dict, optional): the serialized quantization
          configuration. Defaults to None.
    """

    @neural_layer_init(separable=False)
    def __init__(self, *args, quant_config=None, **kwargs):
        # By default neural_layer_init quantizer will be set to -1 (per-axis), but
        # in this very layer it will need to be set per-tensor to complete
        # the conv2d transpose operation. Weight quantizer axis is overridden,
        # and quant_config is updated accordingly.
        self.weight_quantizer.axis = None
        self.quant_config['weight_quantizer']['axis'] = None

        # Limit buffer bitwidth to 27
        self.quant_config['buffer_bitwidth'] = min(28, self.quant_config['buffer_bitwidth'])
        self.buffer_bitwidth = self.quant_config['buffer_bitwidth'] - 1

    @tensor_inputs([FixedPoint])
    @rescale_outputs
    def call(self, inputs):
        # Infer the dynamic output shape
        inputs_shape = tf.shape(inputs)
        out_height = deconv_output_length(
            inputs_shape[1],
            self.kernel_size[0],
            padding=self.padding,
            stride=self.strides[0],
            dilation=self.dilation_rate[0])
        out_width = deconv_output_length(
            inputs_shape[2],
            self.kernel_size[1],
            padding=self.padding,
            stride=self.strides[1],
            dilation=self.dilation_rate[1])
        output_shape = tf.stack((inputs_shape[0], out_height, out_width, 1))
        # Duplicate output_shape to create a placeholder that could be iterated in
        # tf.vectorized_map, making keras happy.
        output_shape = tf.repeat([output_shape], repeats=inputs_shape[-1], axis=0)

        # Quantize the depthwise kernels
        depthwise_kernel = self.weight_quantizer(self.depthwise_kernel)

        # Inputs and kernels must be transposed to have their channel
        # dimension first because the tf.vectorized_map call that follows will
        # unpack them on dimension 0. The channel dimension is virtually
        # restored using expand_dims so that elements have the appropriate
        # shape for the conv2d_transpose call (with a channel dimension of 1
        # which is expected in the depthwise process).
        inputs_channel_first = tf.transpose(inputs, (3, 0, 1, 2))
        inputs_channel_first = tf.expand_dims(inputs_channel_first, -1)
        kernel_channel_first = tf.transpose(depthwise_kernel,
                                            (2, 0, 1, 3))
        kernel_channel_first = tf.expand_dims(kernel_channel_first, -2)

        # Perform the depthwise operation on values using conv2d_transpose on
        # each channel
        dw_values = tf.vectorized_map(
            lambda x: backend.conv2d_transpose(x[0],
                                               x[1],
                                               output_shape=x[2],
                                               strides=self.strides,
                                               padding=self.padding),
            (inputs_channel_first.values, kernel_channel_first.values, output_shape))
        dw_values = tf.transpose(tf.squeeze(dw_values, axis=-1), (1, 2, 3, 0))

        # Last dimension is lost when building layer outputs in model.
        dw_values = tf.reshape(dw_values,
                               (inputs_shape[0], out_height, out_width, inputs_shape[-1]))

        # Build a new FixedPoint
        outputs = FixedPoint(dw_values, inputs.value_bits, inputs.frac_bits)
        # Build a new QFloat
        outputs = QFloat(outputs, kernel_channel_first.scales)

        if self.use_bias:
            # Quantize biases and align them on the outputs
            bias = self.bias_quantizer(self.bias, outputs)
            # Add biases
            outputs = tf.add(outputs, bias)

        return outputs



from .align_rescaling import *
from .invert_batchnorm_pooling import *
from .fold_batchnorms import *
from .remove_reshape import *
from .remove_zeropadding2d import *
from .relu_maxpool import *
from .insert_layer import *
from .replace_lambda import *
from .conv_to_dw import *
from .fold_activations_in_add import *
from .split_layers_with_relu import *
from .sanitize import *



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_pool_ops"]

from onnx import AttributeProto as AP
from onnx.helper import make_node


def get_pool_ops(in_name, out_name, pool_op_type="MaxPool"):
    """Return the pool operation chain.

    Args:
        in_name (str): the input tensor name.
        out_name (str): the required output tensor name.
        pool_op_type (str): the pool type (one of {"MaxPool"}).

    Returns:
        list of NodeProto: the operation chain.
    """
    # Create the node
    if pool_op_type == "MaxPool":
        pool_node = make_node("MaxPool", [in_name], [out_name])
        # Conv have similar attributes than MaxPool. To avoid share them,
        # we define new reference values.
        pool_node.attribute.extend([
            AP(name="kernel_shape", ref_attr_name="pool_size", type=AP.INTS),
            AP(name="strides", ref_attr_name="pool_strides", type=AP.INTS),
            AP(name="pads", ref_attr_name="pool_pads", type=AP.INTS)])
        # Append Maxpool to the operation chain
        nodes = [pool_node]
    elif pool_op_type == "GlobalAvgPool":
        # Append GAP to the operation chain, as ReduceSum
        # Note: spatial dimensions have to fold in output scale
        nodes = [make_node("Constant", [], ["GAP/axes"], value_ints=[-2, -1]),
                 make_node("ReduceSum", [in_name, "GAP/axes"], [out_name])]
    else:
        raise ValueError(f"Unrecognized {pool_op_type} pool operation type.")
    return nodes



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
QuantizedSeparableConv2D layer definition.
"""

__all__ = ["QuantizedSeparableConv2D"]

import tensorflow as tf
import keras

from keras.layers import SeparableConv2D
from keras import backend

from .layers_base import (register_quantize_target, rescale_outputs, tensor_inputs,
                          neural_layer_init, register_aligned_inputs, QuantizedLayer)
from ..tensors import FixedPoint


@register_quantize_target(SeparableConv2D)
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedSeparableConv2D(QuantizedLayer, SeparableConv2D):
    """ A separable convolutional layer that operates on quantized inputs and weights.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """
    arg_constraints = {'depth_multiplier': 1}

    @neural_layer_init(True)
    def __init__(self, *args, quant_config=None, **kwargs):
        pass

    @tensor_inputs([FixedPoint, tf.Tensor])
    @rescale_outputs
    def call(self, inputs):
        # Although the dephwise operation does not require it, we only accept inputs quantized
        # per-tensor to avoid increasing too much the fractional bits of the depthwise outputs.
        inputs.assert_per_tensor()

        # Quantize the weights
        depthwise_kernel = self.dw_weight_quantizer(self.depthwise_kernel)
        pointwise_kernel = self.pw_weight_quantizer(self.pointwise_kernel)

        dw_outputs_q = backend.depthwise_conv2d(
            inputs,
            depthwise_kernel,
            strides=self.strides,
            padding=self.padding,
            dilation_rate=self.dilation_rate,
            data_format=self.data_format)

        outputs = tf.nn.convolution(
            dw_outputs_q,
            pointwise_kernel,
            strides=[1, 1, 1, 1],
            padding='VALID',
            data_format="NHWC" if self.data_format == "channels_last" else "NCHW")

        if self.use_bias:
            # Quantize bias and align it on the outputs
            bias = self.bias_quantizer(self.bias, outputs)
            outputs = tf.add(outputs, bias)

        return outputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Utility methods to insert layers in a model.
"""

__all__ = ['insert_layer', 'insert_rescaling', 'insert_in_config']

from copy import deepcopy

from keras.layers import serialize, InputLayer, Rescaling
from keras.models import Model, Sequential

from .transforms_utils import (get_layer_index, inbound_node_generator,
                               replace_layer_name_for_connection_info, get_layers_by_type)
from ..utils import apply_weights_to_model
from ...layers.quantizers import OutputQuantizer, Dequantizer


def insert_in_config(model, target_layer_name, new_layer, config, outbound_names=None):
    """ Inserts the given layer in the model after the layer with the name target_layer_name by
    editing the given configuration.

    Args:
        model (keras.Model): the model to update
        target_layer_name (str): name of the layer after which to insert a layer
        new_layer (keras.layers.Layer): layer to insert
        config (dict): model dict config being updated
        outbound_names (list, optional): list of outbounds layers names for the inserted
            layer. When not specified, the outbound_names outbounds are retrieved from the given
            model. Providing incoherent names will result in an invalid model graph. Defaults to
            None.
    """
    layers_config = config['layers']

    # Prepare the layer configuration to be inserted
    new_layer_config = serialize(new_layer)

    # Handling sequential and functional models differently:
    #   - sequential models 'layers' configuration is a sorted list of the layers, so we just need
    #     to insert the new layer within that list,
    #   - for functional models, the layers inbound and outbounds are updated first
    if not isinstance(model, Sequential):
        # The layer name is added to the configuration
        new_layer_config['name'] = new_layer.name

        # Retrieve target_layer outbounds if None specified.
        if outbound_names is None:
            target_outbounds = model.get_layer(target_layer_name).outbound_nodes
            outbound_names = [outbound.layer.name for outbound in target_outbounds]

        # OutputQuantizer does not support multiple inputs so target layers with multiple outputs
        # are rejected
        if len(outbound_names) > 1 and isinstance(new_layer, OutputQuantizer):
            raise RuntimeError("Inserting an OutputQuantizer after a layer with multiple outputs "
                               "is not supported.")

        if len(outbound_names):
            # Initialize the new layer inbounds
            new_layer_inbounds = []

            # Replace inbounds from the layers after the target layer with the inserted layer
            outbound_ids = [get_layer_index(layers_config, outbound) for outbound in outbound_names]
            for id in outbound_ids:
                for inbound_node in inbound_node_generator(layers_config[id]):
                    if isinstance(inbound_node, dict):
                        inbound_node = inbound_node.values()
                    for connection_info in inbound_node:
                        matched = replace_layer_name_for_connection_info(connection_info,
                                                                         target_layer_name,
                                                                         new_layer.name)
                        # Store the replaced inbound as it will later be used by the inserted layer
                        if matched and matched not in new_layer_inbounds:
                            new_layer_inbounds.append(matched)

            # Set the inserted layer inbounds
            new_layer_config['inbound_nodes'] = [new_layer_inbounds]

        else:
            # If target layer has no outbounds (ie. it's a model output), update the model
            # output layers list
            for index, out_layer in enumerate(config['output_layers']):
                if out_layer[0] == target_layer_name:
                    config['output_layers'][index][0] = new_layer.name

            # The inserted layer takes the target layer as its inbound
            new_layer_config['inbound_nodes'] = [[[target_layer_name, 0, 0, {}]]]

    # The new layer configuration can now be inserted into the layers config
    layers_config.insert(get_layer_index(layers_config, target_layer_name) + 1, new_layer_config)


def _insert_layer(model, target_layer_name, new_layer):
    """ Inserts the given layer in the model after the layer with the name target_layer_name.

    Args:
        model (keras.Model): the model to update
        target_layer_name (str): name of the layer after which to insert a layer
        new_layer (keras.layers.Layer): layer to insert

    Returns:
        keras.Model: the new model
    """
    # Check that the model has a layer with then given target_layer_name
    if not any(ly.name == target_layer_name for ly in model.layers):
        raise ValueError(f'{target_layer_name} not found in model.')

    # get_config documentation mentions that a copy should be made when planning to modify the
    # config
    config = deepcopy(model.get_config())

    # Insert layer in config graph
    insert_in_config(model, target_layer_name, new_layer, config)

    # Reconstruct model from the config
    if isinstance(model, Sequential):
        new_model = Sequential.from_config(config)
    else:
        new_model = Model.from_config(config)

    # Load original weights
    variables_dict = {var.name: var for var in model.variables}
    apply_weights_to_model(new_model, variables_dict, False)
    return new_model


def insert_layer(model, target_layer_name, new_layer):
    """ Inserts the given layer in the model after the layer with the name target_layer_name.

    Note that new_layer type is restricted to (OutputQuantizer, Dequantizer).

    Args:
        model (keras.Model): the model to update
        target_layer_name (str): name of the layer after which to insert a layer
        new_layer (keras.layers.Layer): layer to insert

    Raises:
        ValueError: when target_layer_name is not found in model or new_layer is not in
            (OutputQuantizer, Dequantizer)

    Returns:
        keras.Model: the new model
    """
    # Check added layer type
    if not isinstance(new_layer, (OutputQuantizer, Dequantizer)):
        raise ValueError(f'Inserted layer must be of type OutputQuantizer or Dequantizer, \
                        `received {type(new_layer)}.')

    return _insert_layer(model, target_layer_name, new_layer)


def insert_rescaling(model, scale, offset):
    """ Inserts a Rescaling as first layer of the Model (after the Input)

    Args:
        model (keras.Model): the model to update
        scale (float): the Rescaling scale
        offset (float): the Rescaling offset

    Raises:
        ValueError: when the Model does not have an Input layer.

    Returns:
        keras.Model: the new model
    """
    first_layers = get_layers_by_type(model, InputLayer)
    if len(first_layers) == 0:
        raise ValueError("Inserting a Rescaling layer in a Model without an Input layer is not"
                         " supported.")
    for target_layer in first_layers:
        model = _insert_layer(model, target_layer.name, Rescaling(scale, offset))
    return model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
quantizeml analysis main command-line interface.
"""

import argparse

from .. import load_model
from .kernel_distribution import plot_kernel_distribution
from .quantization_error_api import (measure_layer_quantization_error,
                                     measure_cumulative_quantization_error)
from .tools import print_metric_table


def add_analysis_arguments(parser):
    asp = parser.add_subparsers(dest="analysis_action")

    # Common arguments
    a_parent_parser = argparse.ArgumentParser(add_help=False)
    a_parent_parser.add_argument("-m", "--model", type=str, required=True, help="Model to analyze")

    # Plot kernel distribution
    k_parser = asp.add_parser("kernel_distribution", parents=[a_parent_parser],
                              help="Plot kernel distribution")
    k_parser.add_argument("-l", "--logdir", type=str, required=True,
                          help="Log directory to save plots")

    # Layer quantization error
    qe_parser = asp.add_parser("quantization_error", parents=[a_parent_parser],
                               help="Measure quantization error")
    qe_parser.add_argument("-qm", "--quantized_model", type=str, required=True,
                           help="The quantized model to analyze")
    qe_parser.add_argument("-tl", "--target_layer", type=str, default=None,
                           help="Compute per_channel error for a specific layer/node. "
                           "Defaults to %(default)s")
    qe_parser.add_argument("-bs", "--batch_size", type=int, default=16,
                           help="Batch size to generate samples. Defaults to %(default)s")
    qe_parser.add_argument("-c", "--cumulative", action="store_true",
                           help="Compute cumulative quantization error instead of isolated one. "
                           "Defaults to %(default)s")


def main(args):
    """ CLI entry point.

    Contains an argument parser with specific arguments to analysis a model.
    Complete arguments lists available using the -h or --help argument.

    """
    model = load_model(args.model)
    if args.analysis_action == "kernel_distribution":
        plot_kernel_distribution(model, logdir=args.logdir)
    elif args.analysis_action == "quantization_error":
        qmodel = load_model(args.quantized_model)
        if args.cumulative:
            summary = measure_cumulative_quantization_error(model,
                                                            qmodel,
                                                            batch_size=args.batch_size,
                                                            target_layer=args.target_layer)
        else:
            summary = measure_layer_quantization_error(model,
                                                       qmodel,
                                                       batch_size=args.batch_size,
                                                       target_layer=args.target_layer)
        model_name = model.name if hasattr(model, "name") else model.graph.name
        print_metric_table(summary, model_name=model_name)
    else:
        raise RuntimeError(f'unknown action: {args.action}')



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_variable"]

import onnx
import onnx.numpy_helper

from onnxruntime.quantization.quant_utils import find_by_name


def get_variable(name, graph):
    """Helper to get the value of an initializar as np.array.

    Args:
        name (str): the name of the variable.
        graph (GraphProto): the graph containing the variable.

    Returns:
        np.array: the value of the variable.
    """
    initializer = find_by_name(name, graph.initializer)
    return onnx.numpy_helper.to_array(initializer)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["BufferTempConv", "DepthwiseBufferTempConv", "reset_buffers",
           "QuantizedBufferTempConv", "QuantizedDepthwiseBufferTempConv"]

import tensorflow as tf
import keras

from keras import layers
from .layers_base import (register_quantize_target, rescale_outputs, tensor_inputs,
                          neural_layer_init, register_aligned_inputs, QuantizedLayer)
from .recorders import (TensorRecorder, NonTrackVariable, NonTrackFixedPointVariable,
                        NonTrackQFloatVariable)
from ..tensors import FixedPoint, QTensor


@keras.saving.register_keras_serializable()
class BaseBufferTempConv(layers.Layer):
    """The base class for BufferTempConv layers.

    Args:
        kernel_size (int): an integer, specifying the height and width of the 1D convolution window.
        use_bias (bool, optional): whether the layer uses a bias vector. Defaults to False
    """

    def __init__(self, kernel_size, use_bias=False, **kwargs):
        super().__init__(**kwargs)
        self.kernel_size = kernel_size
        self.use_bias = use_bias
        self._fifo = NonTrackVariable("fifo")
        self._counter = NonTrackVariable("counter")

    @property
    def fifo(self):
        return getattr(self._fifo, 'var', None)

    @property
    def counter(self):
        return self._counter.var

    def _init_fifo(self, new_sample):
        """Helper function to initialize the fifo variable.

        Args:
            new_sample (tf.Tensor): the new handled input.
        """
        zeros = tf.zeros_like(new_sample)
        init_value = tf.tile(tf.expand_dims(zeros, axis=-2), [1, 1, 1, self.kernel_size, 1])
        # Initialize (only during the first call) the internal state variables
        self._fifo.init_var(init_value)
        # Since the fifo variable will be added to the graph during the layer graph initialization
        # the batch size is unknown and set to None. Such behavior won't allow the next
        # operations due to a conflict in the brodcastable shape. As a workaround, we set again in
        # the variable a tf.zeros tensor but this time with the batch_size known since it's in the
        # call, when counter == 1 (first passage).
        if self.counter == 1:
            self._fifo.set_var(init_value)

    def update_counter(self):
        """Function that increments a 'counter' variable. If the variable doesn't exist, it
            will be created in the layer graph and tracked.
        """
        self._counter.init_var(0, True)
        self._counter.var.assign_add(1)

    def _fifo_op(self, new_sample):
        """Pushes a new sample to the FIFO, popping out the oldest one.

        Args:
            new_sample (tf.Tensor): the new handled input.
        """
        # Update or initialize (if it's the first call) the counter variable
        self.update_counter()

        # Init the fifo
        self._init_fifo(new_sample)

        # FIFO pop operation
        intermediate_fifo = tf.gather(self.fifo, tf.range(1, self.fifo.shape[-2]), axis=-2)
        # FIFO push operation
        self._fifo.set_var(
            tf.concat([intermediate_fifo, tf.expand_dims(new_sample, axis=-2)], axis=-2))

    def _align_fifo(self, fifo_q):
        """Helper to align the layer FIFO and expand it into a higher bitwidth.
        The input_shift is also created if doesn't exist and ready to be stored.

        Args:
            fifo_q (FixedPoint): The input quantized FixedPoint FIFO.

        Returns:
            fifo_aligned: The FixedPoint expanded and aligned fifo.
        """
        # Expand the inputs to a higher bitwidth to avoid saturation and align them.
        fifo_aligned, shift = fifo_q.expand(self.buffer_bitwidth)
        self.input_shift(shift)
        return fifo_aligned

    def call(self, inputs):
        raise NotImplementedError

    def reset_buffer(self):
        """Helper that allows reinitializing the layer internal buffer. it should be called before
        doing any inference on this layer.
        """
        self._fifo.reset_var()
        self._counter.reset_var()

    def get_config(self):
        config = super().get_config()
        config.update({
            'kernel_size': self.kernel_size,
            'use_bias': self.use_bias
        })
        return config


def _buffer_temp_conv_op(fifo, kernel):
    """ BufferTempConv layer main op function.

    Args:
        fifo (tf.Tensor): the remembered inputs stacked in the layer fifo.
        kernel (tf.Tensor): the layer kernel weights.

    Returns:
        output: the layer output tensor.
    """
    # Reshape FIFO to perform a conv2d: (B, H, W, T, C) -> (B, H, W, T * C)
    _, H, W, T, C = fifo.shape
    fifo_reshaped = tf.reshape(fifo, shape=(-1, H, W, T * C))
    # (B, H, W, T * C) * (1, 1, T * C, F) -> (B, H, W, F)
    output = tf.nn.convolution(fifo_reshaped, kernel)
    return output


def _depthwise_buffer_temp_conv_op(fifo, kernel):
    """ DepthwiseBufferTempConv layer main op function.

    Args:
        fifo (tf.Tensor): the remembered inputs stacked in the layer fifo.
        kernel (tf.Tensor): the layer kernel weights.

    Returns:
        output: the layer output tensor.
    """
    # (B, H, W, T, C) * (T, C) -> (B, H, W, T, C)
    output = tf.math.multiply(fifo, kernel)
    # (B, H, W, T, C) -> (B, H, W, C)
    output = tf.math.reduce_sum(output, axis=-2)

    return output


@keras.saving.register_keras_serializable()
class BufferTempConv(BaseBufferTempConv):
    """A Conv3D layer working on the first dimension implemented as a Conv1D associated with a FIFO.
The FIFO buffer acts as a temporal sliding window on inputs for which content is convolved with the
layer kernel.

    Expects the input data to be streamed into the layer sequentially in time.

    Args:
        filters (int): the dimensionality of the output space (i.e. the number of output filters in
            the convolution).
        kernel_size (int): an integer, specifying the height and width of the 1D convolution window.
        use_bias (bool, optional): whether the layer uses a bias vector. Defaults to False
    """

    def __init__(self,
                 filters,
                 kernel_size,
                 use_bias=False,
                 **kwargs):
        super().__init__(kernel_size, use_bias, **kwargs)
        self.filters = filters

    def build(self, input_shape):
        with tf.name_scope(self.name + '/'):
            super().build(input_shape)
            kernel_shape = (1, 1, self.kernel_size * input_shape[-1], self.filters)
            self.kernel = self.add_weight(name="kernel", shape=kernel_shape)
            if self.use_bias:
                self.bias = self.add_weight(name="bias", shape=(self.filters,))

    def call(self, inputs):
        # FIFO pop/push, shape is (B, H, W, T, C)
        self._fifo_op(inputs)
        output = _buffer_temp_conv_op(self.fifo, self.kernel)
        if self.use_bias:
            output = tf.math.add(output, self.bias)
        return output

    def get_config(self):
        config = super().get_config()
        config.update({
            'filters': self.filters
        })
        return config


@keras.saving.register_keras_serializable()
class DepthwiseBufferTempConv(BaseBufferTempConv):
    """Similar to its BufferTempConv sibling, this layer operates with a depthwise kernel.

    Expects the input data to be streamed into the layer sequentially in time.

    Args:
        kernel_size (int): an integer, specifying the height and width of the 1D convolution window.
        use_bias (bool, optional): whether the layer uses a bias vector. Defaults to False
    """

    def build(self, input_shape):
        with tf.name_scope(self.name + '/'):
            super().build(input_shape)
            kernel_shape = (self.kernel_size, input_shape[-1])
            self.kernel = self.add_weight(name="kernel", shape=kernel_shape)
            if self.use_bias:
                self.bias = self.add_weight(name="bias", shape=(input_shape[-1],))

    def call(self, inputs):
        # FIFO pop/push, shape is (B, H, W, T, C)
        self._fifo_op(inputs)

        output = _depthwise_buffer_temp_conv_op(self.fifo, self.kernel)
        if self.use_bias:
            output = tf.math.add(output, self.bias)

        return output


@register_quantize_target([BufferTempConv])
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedBufferTempConv(QuantizedLayer, BufferTempConv):
    """A quantized version of the BufferTempConv layer that operates on quantized inputs and
    weights.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """
    @neural_layer_init(False)
    def __init__(self, *args, quant_config=None, **kwargs):
        self.input_shift = TensorRecorder(name=self.name + "/input_shift")
        self._fifo = None

    @tensor_inputs([QTensor, tf.Tensor])
    @rescale_outputs
    def call(self, inputs):
        if self._fifo is None:
            # Initialize the fifo depending on the input type
            with tf.init_scope():
                self._fifo = NonTrackFixedPointVariable("fifo") if isinstance(inputs, FixedPoint) \
                    else NonTrackQFloatVariable("fifo")

        # FIFO pop/push, shape is (B, H, W, T, C)
        self._fifo_op(inputs)
        # Align FIFO
        fifo_aligned = self._align_fifo(self.fifo)
        # Quantize the weights
        kernel = self.weight_quantizer(self.kernel)
        outputs = _buffer_temp_conv_op(fifo_aligned, kernel)
        if self.use_bias:
            # Quantize bias and align it on the outputs
            bias = self.bias_quantizer(self.bias, outputs)
            outputs = tf.add(outputs, bias)
        return outputs


@register_quantize_target([DepthwiseBufferTempConv])
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedDepthwiseBufferTempConv(QuantizedLayer, DepthwiseBufferTempConv):
    """A quantized version of the DepthwiseBufferTempConv layer that operates on quantized inputs
    and weights.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """
    @neural_layer_init(False)
    def __init__(self, *args, quant_config=None, **kwargs):
        self.input_shift = TensorRecorder(name=self.name + "/input_shift")
        self._fifo = NonTrackFixedPointVariable("fifo")

    @tensor_inputs([FixedPoint])
    @rescale_outputs
    def call(self, inputs):
        # FIFO pop/push, shape is (B, H, W, T, C)
        self._fifo_op(inputs)
        # Align FIFO
        fifo_aligned = self._align_fifo(self._fifo)
        # Quantize the weights
        kernel = self.weight_quantizer(self.kernel)
        outputs = _depthwise_buffer_temp_conv_op(fifo_aligned, kernel)
        if self.use_bias:
            # Quantize bias and align it on the outputs
            bias = self.bias_quantizer(self.bias, outputs)
            outputs = tf.add(outputs, bias)
        return outputs


def reset_buffers(model):
    """ Resets all FIFO-buffer of (Depthwise)BufferTempConv layers within the model.

    Args:
        model (keras.Model): the model to reset
    """
    for layer in model.layers:
        if isinstance(layer, (BufferTempConv, DepthwiseBufferTempConv)):
            layer.reset_buffer()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Transformation to fold unbounded ReLU layers that come after an Add layer.
"""

__all__ = ['fold_activations_in_add']

import keras.layers

from copy import deepcopy
from keras.saving import serialize_keras_object

from ... import layers as quantizeml_layers
from ..utils import apply_weights_to_model
from .transforms_utils import get_layer_index, get_layers, find_layers_pairs, update_inbound


def fold_activations_in_add(model):
    """Returns a new model where Activation layers (Unbounded ReLU) are folded
    into previous Add layers.

    Args:
        model (keras.Model): a model

    Returns:
        keras.Model: the original model or a model with Unbounded ReLU
        layers folded in their preceding Add Layer
    """
    map_add_layer_to_relu = _find_add_relu_to_fold(model)

    # When there are no valid candidates, return the original model
    if len(map_add_layer_to_relu) == 0:
        return model

    model_folded = _fold_activations_in_add(model, map_add_layer_to_relu)

    return model_folded


def _find_add_relu_to_fold(model):
    """Retrieves ReLU layers that can be folded.

    Only ReLUs without a specified maximum value (i.e., unbounded ReLUs)
    that comes after an Add layer are expected to be folded.

    Args:
        model (keras.Model): a model

    Returns:
        dict: map between an Add layer and its ReLU layer to be folded
    """
    supported_layers = (keras.layers.Add, quantizeml_layers.Add)
    add_layer_to_relu_candidates = find_layers_pairs(model, supported_layers, keras.layers.ReLU)

    # Remove pairs that do not meet the condition
    map_add_layer_to_relu = {}
    for inbound, relu in add_layer_to_relu_candidates.items():
        # if the ReLU after the Add has a max_value, we don't fold it
        if relu.max_value is not None:
            continue
        map_add_layer_to_relu[inbound] = relu
    return map_add_layer_to_relu


def _fold_activations_in_add(model, map_add_layer_to_relu):
    """Edits the model configuration to remove ReLU layer and rebuilds a model.

    Args:
        model (keras.Model): a model
        map_add_layer_to_relu (dict): map between an Add layer and its ReLU layer to be folded

    Returns:
        keras.Model: an updated model with Unbounded ReLU layers folded in their preceding Add Layer
    """
    config = deepcopy(model.get_config())
    layers = config['layers']
    new_output_name = None

    for add_layer, relu_layer in map_add_layer_to_relu.items():
        # get index of add layer
        add_index = get_layer_index(layers, add_layer.name)
        name_add = add_layer.name

        # Update Add layer config
        new_config = quantizeml_layers.Add.from_config(layers[add_index]['config'])
        layers[add_index].update(serialize_keras_object(new_config))
        layers[add_index]['config']['activation'] = True

        # Get the layers after relu, ie. outbounds layers
        relu_outbound_names = [outbound.layer.name for outbound in relu_layer.outbound_nodes]
        outbound_ids = [get_layer_index(layers, relu_outbound)
                        for relu_outbound in relu_outbound_names]

        # If ReLU is the last layer (no outbound), store its name and inbound name so that the model
        # output can be updated later
        if len(outbound_ids) == 0:
            new_output_name = name_add
            last_relu = relu_layer.name

        for id in outbound_ids:
            update_inbound(layers[id], relu_layer.name, name_add)

    # Remove ReLU layer
    layers_to_remove = get_layers(config, [relu.name for relu in map_add_layer_to_relu.values()])
    for layer_to_remove in layers_to_remove:
        layers.remove(layer_to_remove)

    # Update the model outputs if needed
    if new_output_name:
        for index, out_layer in enumerate(config['output_layers']):
            if out_layer[0] == last_relu:
                config['output_layers'][index][0] = new_output_name

    updated_model = model.from_config(config)
    variables_dict = {var.name: var for var in model.variables}
    apply_weights_to_model(updated_model, variables_dict)

    return updated_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_input_shift_ops"]

from onnx.helper import make_node


def get_input_shift_ops(in_name, s_name, out_name):
    """Return the input shift operation chain

    Args:
        in_name (str): the input tensor name.
        s_name (str): the shift tensor name.
        out_name (str): the required output tensor name.

    Returns:
        list of NodeProto: the operation chain.
    """
    return [make_node("Mul", inputs=[in_name, s_name], outputs=[out_name])]



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["CalibrationDataReader"]

import numpy as np
import warnings

import onnx
from onnxruntime.quantization import CalibrationDataReader as _CalibrationDataReader

from .model import ONNXModel


def gen_random_samples(input_shape, dtype="float32"):
    output_type = np.dtype(dtype)
    rng = np.random.default_rng()
    if issubclass(output_type.type, np.integer):
        iinfo = np.iinfo(output_type)
        return rng.integers(iinfo.min, iinfo.max, input_shape, dtype=dtype)
    return rng.uniform(-1.0, 1.0, input_shape).astype(dtype)


class CalibrationDataReader(_CalibrationDataReader):
    """Object to read or generate a set of samples to calibrate an ONNX model to be quantized.

    If samples are not specified, generate random samples in the range of [-1, 1]
    when input model type is float, otherwise infer ranges.

    Common use mode:
    >>> dr = CalibrationDataReader(onnx_path, num_samples=10, batch_size=1)
    >>> sample = dr.get_next()
    >>> assert sample[dr.inputs_name].shape[0] == 1
    >>> assert sample[dr.inputs_name].min() >= -1
    >>> assert sample[dr.inputs_name].max() <= 1

    Args:
        model (str or ModelProto or ONNXModel): the ONNX model (or its path) to be calibrated.
        samples (str or np.ndarray, optional): the samples (or its path) to process.
            If not provided, generate random samples following the model input shape
            and the batch_size attribute. Defaults to None.
        num_samples (int, optional): the number of samples to generate.
            Ignore it if samples are provided. Defaults to None.
        batch_size (int, optional): split samples in batches.
            Overwrite it when the model has static inputs. Defaults to 1.
    """

    def __init__(self,
                 model,
                 samples=None,
                 num_samples=None,
                 batch_size=1):

        # Read model
        model = _read_model(model)
        self.input_name = model.input[0].name
        input_shape = model.get_input_shape(input_name=self.input_name)
        input_type = model.get_input_dtype(input_name=self.input_name)

        # Read/Generate dataset
        self.batch_size = input_shape[0] if isinstance(input_shape[0], int) else batch_size or 1
        self.dataset = _read_dataset(samples, input_shape, input_type, num_samples=num_samples)

        # Truncate samples to fit model batch size (when is static)
        if isinstance(input_shape[0], int):
            N, res = divmod(self.dataset.shape[0], self.batch_size)
            N *= self.batch_size
            if res != 0:
                warnings.warn("Truncating samples to fit model batch size "
                              f"({N} instead of {self.dataset.shape[0]}).")
                self.dataset = self.dataset[:N]

        # Process samples
        self.index = 0
        self.num_samples = self.dataset.shape[0] / self.batch_size

    def get_next(self):
        if self.index >= self.num_samples:
            print(f"\rCalibrating with {self.index}/{self.num_samples} samples", end="")
            return print()

        sample = self.dataset[self.index * self.batch_size:(self.index + 1) * self.batch_size]
        self.index += 1
        return {self.input_name: sample}

    def rewind(self):
        self.index = 0


def _read_dataset(samples, input_shape, input_type, num_samples=None):
    # Parse sample
    if isinstance(samples, str):
        data = np.load(samples)
        samples = np.concatenate([data[item] for item in data.files])
    elif samples is None:
        if num_samples is None:
            raise ValueError("Either samples or num_samples must be specified")
        samples = gen_random_samples((num_samples, *input_shape[1:]), input_type)
    if not isinstance(samples, np.ndarray):
        raise ValueError(f"Unrecognized '{type(samples)}' samples.")

    # Check when samples comes from file
    assert samples.dtype == input_type, "Samples type does not match with model input one."
    if samples.shape[1:] != input_shape[1:]:
        raise RuntimeError("Samples shape does not match model input shape. "
                           f"Please verify samples are compatible with {input_shape}.")
    return samples


def _read_model(model):
    if isinstance(model, str):
        model = onnx.load_model(model)
    if isinstance(model, onnx.ModelProto):
        model = ONNXModel(model)
    if not isinstance(model, ONNXModel):
        raise ValueError(f"Unrecognized '{type(model)}' model.")

    assert len(model.input) == 1, "multi-input are not supported models yet"
    assert len(model.output) == 1, "multi-outputs are not supported models yet"
    return model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Helper that replaces Conv2D layers to DepthwiseConv2D when they behave as the latest.
"""

__all__ = ["convert_conv_to_dw_conv"]

import numpy as np
from copy import deepcopy
from keras.layers import Conv2D

from .transforms_utils import get_layers_by_type, get_layer_index


def convert_conv_to_dw_conv(model):
    """ Replaces Conv2D layers from a model to DepthwiseConv2D layers
    if groups==filters==input_channels, since they will have the same behavior.

    Args:
        model (keras.Model): the model of interest

    Returns:
        keras.Model: the original model or the modified one.
    """

    # Get all Conv2D layers present in the model
    conv2ds = get_layers_by_type(model, Conv2D)

    # When there are no valid candidates, return the original model
    if not conv2ds:
        return model

    # Copy configuration before applying modifications
    config = deepcopy(model.get_config())

    convertible_conv2d_names = list()
    for conv2d in conv2ds:
        layer_index = get_layer_index(config['layers'], conv2d.name)
        conv_config = config['layers'][layer_index]
        # Set useful param alias
        conv_input_channel = conv2d.input_shape[-1]
        conv_groups = conv_config['config']['groups']
        conv_filters = conv_config['config']['filters']
        # If in_channel==groups==filters the layer behaves as a DepthwiseConv2D
        if conv_input_channel == conv_groups == conv_filters:
            # keep track of convertible Conv2D layers
            convertible_conv2d_names.append(conv2d.name)
            conv_config['class_name'] = 'DepthwiseConv2D'
            for param in ['kernel_regularizer', 'kernel_initializer', 'kernel_constraint']:
                conv_config['config'][param.replace(
                    'kernel', 'depthwise')] = conv_config['config'].pop(param)
            for param in ['filters', 'groups']:
                conv_config['config'].pop(param)
    # Reconstruct model from the config
    updated_model = model.from_config(config)

    # Restore model weights
    for layer in updated_model.layers:
        weights = model.get_layer(layer.name).get_weights()
        # In order to set Conv2D (depthwise like) kernel into the newly created equivalent
        # DepthwiseConv2D, the last two dims of the kernel should be transposed.
        if layer.name in convertible_conv2d_names:
            assert weights[0].ndim == 4, "the layer kernel must have 4 dims"
            assert weights[0].shape[2] == 1, "The kernel 3rd dim must be 1"
            # Depthwise spatial kernels: ((H, W, C=1, F)) -> (H, W, F, 1)
            weights[0] = np.transpose(weights[0], axes=[0, 1, 3, 2])
        layer.set_weights(weights)

    return updated_model
